from math import isclose, pi
from typing import Any, Dict, List, cast

import pytest
from hypothesis import given
from hypothesis import strategies as st

from opentrons_shared_data.labware.labware_definition import (
    ConicalFrustum,
    CuboidalFrustum,
    HeightVolumePair,
    InnerWellGeometry,
    SphericalSegment,
    UserDefinedVolumes,
)

from opentrons.protocol_engine.errors.exceptions import InvalidLiquidHeightFound
from opentrons.protocol_engine.state.inner_well_math_utils import (
    _cross_section_area_circular,
    _cross_section_area_rectangular,
    _get_segment_capacity,
    _height_from_volume_circular,
    _height_from_volume_rectangular,
    _height_from_volume_spherical,
    _rectangular_frustum_polynomial_roots,
    _reject_unacceptable_heights,
    _volume_from_height_circular,
    _volume_from_height_rectangular,
    _volume_from_height_spherical,
    find_height_inner_well_geometry,
    find_height_user_defined_volumes,
    find_volume_inner_well_geometry,
    find_volume_user_defined_volumes,
)


@pytest.fixture
def user_defined_volumes_params() -> Dict[str, Any]:
    """Return a UserDefinedVolumes BaseModel."""
    params = {}
    params["obj"] = UserDefinedVolumes(
        heightToVolumeMap=[
            HeightVolumePair(height=0.4, volume=2.0),
            HeightVolumePair(height=2.3, volume=9.8),
            HeightVolumePair(height=4.5, volume=12.2),
            HeightVolumePair(height=7.8, volume=50.1),
        ]
    )
    params["volume_inputs_expected_outputs"] = [  # type: ignore[assignment]
        (0.2, 1.0),
        (2.1, 8.9789),
        (2.5, 10.01818),
        (4.5, 12.2),
        (6.0, 29.42727),
    ]
    params["height_inputs_expected_outputs"] = [  # type: ignore[assignment]
        (0.4, 0.08),
        (5.5, 1.2525),
        (9.8, 2.3),
        (0.0, 0.0),
        (50.1, 7.8),
        (40.0, 6.92),
    ]
    return params


def fake_frusta() -> List[List[Any]]:
    """A bunch of weird fake well shapes."""
    frusta = []
    frusta.append(
        [
            CuboidalFrustum(
                shape="cuboidal",
                topXDimension=9.0,
                topYDimension=10.0,
                bottomXDimension=8.0,
                bottomYDimension=9.0,
                topHeight=10.0,
                bottomHeight=5.0,
            ),
            CuboidalFrustum(
                shape="cuboidal",
                topXDimension=8.0,
                topYDimension=9.0,
                bottomXDimension=15.0,
                bottomYDimension=18.0,
                topHeight=5.0,
                bottomHeight=2.0,
            ),
            ConicalFrustum(
                shape="conical",
                topDiameter=23.0,
                bottomDiameter=3.0,
                topHeight=2.0,
                bottomHeight=1.0,
            ),
            SphericalSegment(
                shape="spherical",
                radiusOfCurvature=4.0,
                topHeight=1.0,
                bottomHeight=0.0,
            ),
        ]
    )
    frusta.append(
        [
            CuboidalFrustum(
                shape="cuboidal",
                topXDimension=8.0,
                topYDimension=70.0,
                bottomXDimension=7.0,
                bottomYDimension=75.0,
                topHeight=3.5,
                bottomHeight=2.0,
            ),
            CuboidalFrustum(
                shape="cuboidal",
                topXDimension=8.0,
                topYDimension=80.0,
                bottomXDimension=8.0,
                bottomYDimension=90.0,
                topHeight=1.0,
                bottomHeight=0.0,
            ),
        ]
    )
    frusta.append(
        [
            ConicalFrustum(
                shape="conical",
                topDiameter=23.0,
                bottomDiameter=11.5,
                topHeight=7.5,
                bottomHeight=5.0,
            ),
            ConicalFrustum(
                shape="conical",
                topDiameter=11.5,
                bottomDiameter=23.0,
                topHeight=5.0,
                bottomHeight=2.5,
            ),
            ConicalFrustum(
                shape="conical",
                topDiameter=23.0,
                bottomDiameter=11.5,
                topHeight=2.5,
                bottomHeight=0.0,
            ),
        ]
    )
    frusta.append(
        [
            ConicalFrustum(
                shape="conical",
                topDiameter=4.0,
                bottomDiameter=5.0,
                topHeight=3.0,
                bottomHeight=2.0,
            ),
            SphericalSegment(
                shape="spherical",
                radiusOfCurvature=3.5,
                topHeight=2.0,
                bottomHeight=0.0,
            ),
        ]
    )
    frusta.append(
        [
            SphericalSegment(
                shape="spherical",
                radiusOfCurvature=4.0,
                topHeight=3.0,
                bottomHeight=0.0,
            )
        ]
    )
    frusta.append(
        [
            CuboidalFrustum(
                shape="cuboidal",
                topXDimension=27.0,
                topYDimension=36.0,
                bottomXDimension=36.0,
                bottomYDimension=26.0,
                topHeight=3.5,
                bottomHeight=1.5,
            ),
            SphericalSegment(
                shape="spherical",
                radiusOfCurvature=4.0,
                topHeight=1.5,
                bottomHeight=0.0,
            ),
        ]
    )
    return frusta


@pytest.mark.parametrize(
    ["max_height", "potential_heights", "expected_heights"],
    [
        (34, [complex(4, 5), complex(5, 0), 35, 34, 33, 10, 0], [5, 34, 33, 10, 0]),
        (2934, [complex(4, 5), complex(5, 0)], [5]),
        (100, [-99, -1, complex(99.99, 0), 101], [99.99]),
        (2, [0, -1, complex(-1.5, 0)], [0]),
        (8, [complex(7, 1), -0.01], []),
    ],
)
def test_reject_unacceptable_heights(
    max_height: float, potential_heights: List[Any], expected_heights: List[float]
) -> None:
    """Make sure we reject all mathematical solutions that are physically not possible."""
    if len(expected_heights) != 1:
        with pytest.raises(InvalidLiquidHeightFound):
            _reject_unacceptable_heights(
                max_height=max_height, potential_heights=potential_heights
            )
    else:
        found_heights = _reject_unacceptable_heights(
            max_height=max_height, potential_heights=potential_heights
        )
        assert found_heights == expected_heights[0]


@pytest.mark.parametrize("diameter", [2, 5, 8, 356, 1000])
def test_cross_section_area_circular(diameter: float) -> None:
    """Test circular area calculation."""
    expected_area = pi * (diameter / 2) ** 2
    assert _cross_section_area_circular(diameter) == expected_area


@pytest.mark.parametrize(
    ["x_dimension", "y_dimension"], [(1, 38402), (234, 983), (94857, 40), (234, 999)]
)
def test_cross_section_area_rectangular(x_dimension: float, y_dimension: float) -> None:
    """Test rectangular area calculation."""
    expected_area = x_dimension * y_dimension
    assert (
        _cross_section_area_rectangular(
            x_dimension=x_dimension, y_dimension=y_dimension
        )
        == expected_area
    )


@pytest.mark.parametrize("well", fake_frusta())
@given(target_height_st=st.data())
def test_volume_and_height_circular(well: List[Any], target_height_st: Any) -> None:
    """Test both volume and height calculations for circular frusta."""
    if well[-1].shape == "spherical":
        return
    if any([seg.shape != "conical" for seg in well]):
        return
    for segment in well:
        if segment.shape == "conical":
            a = segment.topDiameter / 2
            b = segment.bottomDiameter / 2
            # test volume within a bunch of arbitrary heights
            segment_height = segment.topHeight - segment.bottomHeight
            for i in range(50):
                target_height = target_height_st.draw(
                    st.floats(
                        min_value=0,
                        max_value=segment_height,
                        allow_infinity=False,
                        allow_nan=False,
                        width=32,
                    )
                )
                r_y = (target_height / segment_height) * (a - b) + b
                expected_volume = (pi * target_height / 3) * (b**2 + b * r_y + r_y**2)
                found_volume = _volume_from_height_circular(
                    target_height=target_height,
                    segment=segment,
                )
                assert isclose(found_volume, expected_volume)
                # test going backwards to get height back
                found_height = _height_from_volume_circular(
                    target_volume=found_volume, segment=segment
                )
                assert isclose(found_height, target_height, abs_tol=0.001)


@pytest.mark.parametrize("well", fake_frusta())
def test_volume_and_height_rectangular(well: List[Any]) -> None:
    """Test both volume and height calculations for rectangular frusta."""
    if well[-1].shape == "spherical":
        return
    total_height = well[0].topHeight
    for segment in well:
        if segment.shape == "cuboidal":
            top_length = segment.topYDimension
            top_width = segment.topXDimension
            bottom_length = segment.bottomYDimension
            bottom_width = segment.bottomXDimension
            a = (
                (top_length - bottom_length)
                * (top_width - bottom_width)
                / (3 * total_height**2)
            )
            b = (
                (bottom_length * (top_width - bottom_width))
                + (bottom_width * (top_length - bottom_length))
            ) / (2 * total_height)
            c = bottom_length * bottom_width
            assert _rectangular_frustum_polynomial_roots(
                top_length=top_length,
                bottom_length=bottom_length,
                top_width=top_width,
                bottom_width=bottom_width,
                total_frustum_height=total_height,
            ) == (a, b, c)
            # test volume within a bunch of arbitrary heights
            for target_height in range(round(total_height)):
                expected_volume = (
                    a * (target_height**3) + b * (target_height**2) + c * target_height
                )
                found_volume = _volume_from_height_rectangular(
                    target_height=target_height,
                    total_frustum_height=total_height,
                    bottom_length=bottom_length,
                    bottom_width=bottom_width,
                    top_length=top_length,
                    top_width=top_width,
                )
                assert found_volume == expected_volume
                # test going backwards to get height back
                found_height = _height_from_volume_rectangular(
                    volume=found_volume,
                    total_frustum_height=total_height,
                    bottom_length=bottom_length,
                    bottom_width=bottom_width,
                    top_length=top_length,
                    top_width=top_width,
                )
                assert isclose(found_height, target_height)


@pytest.mark.parametrize("well", fake_frusta())
def test_volume_and_height_spherical(well: List[Any]) -> None:
    """Test both volume and height calculations for spherical segments."""
    if well[0].shape == "spherical":
        for target_height in range(round(well[0].topHeight)):
            expected_volume = (
                (1 / 3)
                * pi
                * (target_height**2)
                * (3 * well[0].radiusOfCurvature - target_height)
            )
            found_volume = _volume_from_height_spherical(
                target_height=target_height,
                radius_of_curvature=well[0].radiusOfCurvature,
            )
            assert found_volume == expected_volume
            found_height = _height_from_volume_spherical(
                volume=found_volume,
                radius_of_curvature=well[0].radiusOfCurvature,
                total_frustum_height=well[0].topHeight,
            )
            assert isclose(found_height, target_height)


@pytest.mark.parametrize("well", fake_frusta())
def test_height_at_volume_at_section_boundaries(well: List[Any]) -> None:
    """Test that finding the height when volume 0 or ~= capacity  works."""
    inner_well_geometry = InnerWellGeometry(sections=well)
    sorted_well = sorted(
        inner_well_geometry.sections, key=lambda section: section.topHeight
    )
    running_volume = 0.0
    height = find_height_inner_well_geometry(
        target_volume=0.0, well_geometry=inner_well_geometry
    )
    assert isinstance(height, float)
    assert isclose(height, 0.0)
    for segment in sorted_well:
        running_volume += _get_segment_capacity(segment)
        height = find_height_inner_well_geometry(
            target_volume=running_volume,
            well_geometry=inner_well_geometry,
        )
        assert isinstance(height, float)
        assert isclose(height, segment.topHeight)


@pytest.mark.parametrize("well", fake_frusta())
def test_volume_at_section_boundary_heights(well: List[Any]) -> None:
    """Test that finds the volume at the segment boundaries (top/bottom)."""
    inner_well_geometry = InnerWellGeometry(sections=well)
    tot_ul = 0.0
    # reverse b/c list of top->bottom
    for segment in reversed(well):
        bottom_ul = find_volume_inner_well_geometry(
            target_height=segment.bottomHeight, well_geometry=inner_well_geometry
        )
        assert isclose(cast(float, bottom_ul), tot_ul)
        top_ul = find_volume_inner_well_geometry(
            target_height=segment.topHeight, well_geometry=inner_well_geometry
        )
        tot_ul += _get_segment_capacity(segment)
        assert isclose(cast(float, top_ul), tot_ul)


def test_user_volumes_raises_error_for_invalid_target(
    user_defined_volumes_params: Dict[str, Any],
) -> None:
    """Test that UserDefinedVolumes calculations reject target inputs that are not allowed."""
    user_defined_volumes_obj = user_defined_volumes_params["obj"]
    max_defined_height = user_defined_volumes_obj.heightToVolumeMap[-1].height
    max_defined_volume = user_defined_volumes_obj.heightToVolumeMap[-1].volume
    min_defined_height = user_defined_volumes_obj.heightToVolumeMap[0].height
    min_defined_volume = user_defined_volumes_obj.heightToVolumeMap[0].volume

    # less than 0 should raise an error
    with pytest.raises(InvalidLiquidHeightFound):
        find_volume_user_defined_volumes(
            target_height=-0.01, well_geometry=user_defined_volumes_obj
        )
    with pytest.raises(InvalidLiquidHeightFound):
        find_height_user_defined_volumes(
            target_volume=-0.01, well_geometry=user_defined_volumes_obj
        )

    # between 0 and min defined height should be ok
    vol = find_volume_user_defined_volumes(
        target_height=(min_defined_height / 2), well_geometry=user_defined_volumes_obj
    )
    assert vol is not None
    height = find_height_user_defined_volumes(
        target_volume=(min_defined_volume / 2), well_geometry=user_defined_volumes_obj
    )
    assert height is not None

    # betwen min defined height and max defined height should be ok
    vol = find_volume_user_defined_volumes(
        target_height=((min_defined_height + max_defined_height) / 2),
        well_geometry=user_defined_volumes_obj,
    )
    assert vol is not None
    height = find_height_user_defined_volumes(
        target_volume=((min_defined_volume + max_defined_volume) / 2),
        well_geometry=user_defined_volumes_obj,
    )
    assert height is not None

    # any greater than max defined height should cause an error
    with pytest.raises(InvalidLiquidHeightFound):
        find_volume_user_defined_volumes(
            target_height=max_defined_height + 0.01,
            well_geometry=user_defined_volumes_obj,
        )
    with pytest.raises(InvalidLiquidHeightFound):
        find_height_user_defined_volumes(
            target_volume=max_defined_volume + 0.01,
            well_geometry=user_defined_volumes_obj,
        )


def test_get_user_volumes(user_defined_volumes_params: Dict[str, Any]) -> None:
    """Test linear interpolation math for user-defined volumes."""
    user_defined_volumes_obj = user_defined_volumes_params["obj"]
    inputs_expected_outputs = user_defined_volumes_params[
        "volume_inputs_expected_outputs"
    ]
    for height, expected_vol in inputs_expected_outputs:
        volume_estimate = find_volume_user_defined_volumes(
            target_height=height, well_geometry=user_defined_volumes_obj
        )
        assert isinstance(volume_estimate, float)
        assert isclose(volume_estimate, expected_vol, abs_tol=0.001)


def test_get_user_heights(
    user_defined_volumes_params: Dict[str, Any],
) -> None:
    """Test linear interpolation math for user-defined volumes."""
    user_defined_volumes_obj = user_defined_volumes_params["obj"]
    inputs_expected_outputs = user_defined_volumes_params[
        "height_inputs_expected_outputs"
    ]

    for vol, expected_height in inputs_expected_outputs:
        height_estimate = find_height_user_defined_volumes(
            target_volume=vol, well_geometry=user_defined_volumes_obj
        )
        assert isinstance(height_estimate, float)
        assert isclose(height_estimate, expected_height, abs_tol=0.001)



"""Test retractAxis command."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.retract_axis import (
    RetractAxisImplementation,
    RetractAxisParams,
    RetractAxisResult,
)
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.types import MotorAxis


async def test_retract_axis_implementation(
    decoy: Decoy,
    movement: MovementHandler,
) -> None:
    """A retractAxis command should have an execution implementation."""
    subject = RetractAxisImplementation(movement=movement)

    data = RetractAxisParams(axis=MotorAxis.Y)
    result = await subject.execute(data)

    assert result == SuccessData(
        public=RetractAxisResult(),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )
    decoy.verify(await movement.retract_axis(axis=MotorAxis.Y))



"""Tests for opentrons.protocol_reader.file_reader_writer.FileReaderWriter."""

from pathlib import Path

import pytest

from ._input_file import InputFile
from opentrons.protocol_reader.file_reader_writer import (
    BufferedFile,
    FileReadError,
    FileReaderWriter,
)


async def test_read_with_abstract_input_file() -> None:
    """It should read file-likes."""
    file_1 = InputFile.make(filename="hello.txt", contents=b"# hello")
    file_2 = InputFile.make(filename="world.txt", contents=b"# world")

    subject = FileReaderWriter()
    result = await subject.read([file_1, file_2])

    assert result == [
        BufferedFile(name="hello.txt", contents=b"# hello", path=None),
        BufferedFile(name="world.txt", contents=b"# world", path=None),
    ]


async def test_read_with_path(tmp_path: Path) -> None:
    """It should read paths."""
    path_1 = tmp_path / "hello.txt"
    path_2 = tmp_path / "world.txt"

    path_1.write_text("# hello")
    path_2.write_text("# world")

    subject = FileReaderWriter()
    result = await subject.read([path_1, path_2])

    assert result == [
        BufferedFile(name="hello.txt", contents=b"# hello", path=path_1),
        BufferedFile(name="world.txt", contents=b"# world", path=path_2),
    ]


async def test_read_missing_filename() -> None:
    """It should error if a file has no filename."""
    in_file = InputFile.make(filename="", contents=b"")

    subject = FileReaderWriter()

    with pytest.raises(FileReadError, match="missing a name"):
        await subject.read([in_file])


async def test_write(tmp_path: Path) -> None:
    """It should write buffered files to disk."""
    directory = tmp_path / "target"
    file_1 = BufferedFile(name="hello.txt", contents=b"# hello", path=None)
    file_2 = BufferedFile(name="world.txt", contents=b"# world", path=None)

    subject = FileReaderWriter()
    await subject.write(
        directory=directory,
        files=[file_1, file_2],
    )

    assert Path(tmp_path / "target" / "hello.txt").read_bytes() == b"# hello"
    assert Path(tmp_path / "target" / "world.txt").read_bytes() == b"# world"



"""Tests for the command QueueWorker in opentrons.protocol_engine."""

from typing import AsyncGenerator, Callable, Generator

import pytest
from decoy import Decoy, matchers

from opentrons.protocol_engine.errors import RunStoppedError
from opentrons.protocol_engine.execution import CommandExecutor, QueueWorker
from opentrons.protocol_engine.state.state import StateStore


class BreakLoopError(Exception):
    """An exception to break out of the worker's wait-for-new-command loop."""


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def command_executor(decoy: Decoy) -> CommandExecutor:
    """Get a mocked out CommandExecutor."""
    return decoy.mock(cls=CommandExecutor)


@pytest.fixture
def command_generator(decoy: Decoy) -> Callable[[], AsyncGenerator[str, None]]:
    """Get a mocked out CommandExecutor."""

    async def generator() -> AsyncGenerator[str, None]:
        yield "command-id-1"
        yield "command-id-2"
        yield "command-id-3"

    return generator


@pytest.fixture
def subject(
    state_store: StateStore,
    command_executor: CommandExecutor,
    command_generator: Callable[[], AsyncGenerator[str, None]],
) -> QueueWorker:
    """Get a QueueWorker instance."""
    return QueueWorker(
        state_store=state_store,
        command_executor=command_executor,
        command_generator=command_generator,
    )


@pytest.fixture(autouse=True)
async def queue_commands(decoy: Decoy, state_store: StateStore) -> None:
    """Load the command queue with 2 queued commands, then stop.

    When state_store.wait_for(...) is called, return "command-id-1" the first time,
    return "command-id-2" the second time, and raise RunStoppedError the third time.
    """

    def get_next_to_execute() -> Generator[str, None, None]:
        yield "command-id-1"
        yield "command-id-2"
        raise RunStoppedError()

    get_next_to_execute_results = get_next_to_execute()

    decoy.when(
        await state_store.wait_for(condition=state_store.commands.get_next_to_execute)
    ).then_do(lambda *args, **kwargs: next(get_next_to_execute_results))


async def test_start_processes_commands(
    decoy: Decoy,
    state_store: StateStore,
    command_executor: CommandExecutor,
    subject: QueueWorker,
) -> None:
    """It should pull commands off the queue and execute them."""
    subject.start()

    decoy.verify(
        await command_executor.execute(command_id="command-id-1"),
        times=0,
    )

    await subject.join()

    decoy.verify(
        await command_executor.execute(command_id="command-id-1"),
        await command_executor.execute(command_id="command-id-2"),
    )


async def test_cancel(
    decoy: Decoy,
    state_store: StateStore,
    command_executor: CommandExecutor,
    subject: QueueWorker,
) -> None:
    """It should stop pulling jobs if it is cancelled."""
    subject.start()
    subject.cancel()
    await subject.join()

    decoy.verify(
        await command_executor.execute(command_id=matchers.Anything()),
        times=0,
    )
    decoy.verify(command_executor.cancel_tasks("Engine cancelled"), times=1)


async def test_cancel_noops_if_joined(
    decoy: Decoy,
    state_store: StateStore,
    command_executor: CommandExecutor,
    subject: QueueWorker,
) -> None:
    """It should noop on cancel if the worker has already been `join`'d."""
    subject.start()
    await subject.join()
    subject.cancel()
    decoy.verify(command_executor.cancel_tasks("Engine commands complete"), times=1)


async def test_unhandled_exception_breaks_loop(
    decoy: Decoy,
    state_store: StateStore,
    command_executor: CommandExecutor,
    subject: QueueWorker,
) -> None:
    """It should raise any unhandled exceptions in `join`."""
    decoy.when(await command_executor.execute(command_id="command-id-1")).then_raise(  # type: ignore[func-returns-value]
        RuntimeError("oh no")
    )

    subject.start()

    with pytest.raises(RuntimeError, match="oh no"):
        await subject.join()
    decoy.verify(command_executor.cancel_tasks("Engine failed"), times=1)


async def test_engine_stopped_exception_breaks_loop_gracefully(
    decoy: Decoy,
    state_store: StateStore,
    command_executor: CommandExecutor,
    subject: QueueWorker,
) -> None:
    """It should `join` gracefully if a RunStoppedError is raised."""
    decoy.when(
        await state_store.wait_for(condition=state_store.commands.get_next_to_execute)
    ).then_raise(RunStoppedError("oh no"))

    subject.start()
    await subject.join()
    decoy.verify(command_executor.cancel_tasks("Engine commands complete"), times=1)



"""Test pause command."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.wait_for_resume import (
    WaitForResumeCreate,
    WaitForResumeImplementation,
    WaitForResumeParams,
    WaitForResumeResult,
)
from opentrons.protocol_engine.execution import RunControlHandler


async def test_wait_for_resume_implementation(
    decoy: Decoy,
    run_control: RunControlHandler,
) -> None:
    """It should wait a resume from the RunControlHandler."""
    subject = WaitForResumeImplementation(run_control=run_control)

    data = WaitForResumeParams(message="hello world")

    result = await subject.execute(data)

    assert result == SuccessData(public=WaitForResumeResult())
    decoy.verify(await run_control.wait_for_resume(), times=1)


def test_wait_for_resume_accepts_pause_command_type() -> None:
    """It should accept commandType=pause for backwards compatibility."""
    result = WaitForResumeCreate(commandType="pause", params=WaitForResumeParams())
    assert result.commandType == "pause"



import pytest

from opentrons_shared_data.deck.types import DeckDefinitionV3

from opentrons import protocol_api as papi
from opentrons import types

labware_name = "corning_96_wellplate_360ul_flat"


# labware._core.get_geometry() is an implementation detail and is not present in ProtocolContexts
# backed by Protocol Engine.
# TODO(mm, 2022-04-28): Make sure this logic is tested elsewhere, then delete this test.
@pytest.mark.apiv2_non_pe_only
def test_load_to_slot(
    ctx: papi.ProtocolContext, legacy_deck_definition: DeckDefinitionV3
) -> None:
    slot_1 = [
        slot
        for slot in legacy_deck_definition["locations"]["orderedSlots"]
        if slot["id"] == "1"
    ][0]
    slot_2 = [
        slot
        for slot in legacy_deck_definition["locations"]["orderedSlots"]
        if slot["id"] == "2"
    ][0]
    labware = ctx.load_labware(labware_name, "1")

    assert labware._core.get_geometry().offset == types.Point(*slot_1["position"])  # type: ignore[attr-defined]
    other = ctx.load_labware(labware_name, 2)
    assert other._core.get_geometry().offset == types.Point(*slot_2["position"])  # type: ignore[attr-defined]


def test_loaded(ctx: papi.ProtocolContext) -> None:
    labware = ctx.load_labware(labware_name, "1")
    assert ctx.loaded_labwares[1] == labware


def test_get_incorrect_definition_by_name() -> None:
    with pytest.raises(FileNotFoundError):
        papi.labware.get_labware_definition("fake_labware")


def test_get_mixed_case_labware_def() -> None:
    dfn = papi.labware.get_labware_definition("COrnIng_96_wElLplaTE_360UL_Flat")
    assert dfn["parameters"]["loadName"] == labware_name


def test_load_label(ctx: papi.ProtocolContext) -> None:
    labware = ctx.load_labware(labware_name, "1", "my cool labware")
    assert "my cool labware" in str(labware)
    assert labware._core.get_user_display_name() == "my cool labware"


def test_deprecated_load(ctx: papi.ProtocolContext) -> None:
    labware = ctx.load_labware_by_name(labware_name, "1", "my cool labware")
    assert "my cool labware" in str(labware)



"""Test Temperature Module's await temperature command implementation."""

from decoy import Decoy

from opentrons.hardware_control.modules import TempDeck
from opentrons.protocol_engine.commands import temperature_module
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.temperature_module.deactivate import (
    DeactivateTemperatureImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    TemperatureModuleId,
    TemperatureModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView


async def test_await_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should be able to wait for the module's target temperature."""
    subject = DeactivateTemperatureImpl(state_view=state_view, equipment=equipment)

    data = temperature_module.DeactivateTemperatureParams(moduleId="tempdeck-id")

    module_substate = decoy.mock(cls=TemperatureModuleSubState)
    tempdeck_hardware = decoy.mock(cls=TempDeck)

    decoy.when(
        state_view.modules.get_temperature_module_substate(module_id="tempdeck-id")
    ).then_return(module_substate)

    decoy.when(module_substate.module_id).then_return(
        TemperatureModuleId("tempdeck-id")
    )

    # Get stubbed hardware module
    decoy.when(
        equipment.get_module_hardware_api(TemperatureModuleId("tempdeck-id"))
    ).then_return(tempdeck_hardware)

    result = await subject.execute(data)
    decoy.verify(await tempdeck_hardware.deactivate(), times=1)
    assert result == SuccessData(
        public=temperature_module.DeactivateTemperatureResult(),
    )



"""Tests for liquid class blowout properties in the Opentrons protocol API."""

from typing import Any

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st
from pydantic import ValidationError

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    BlowoutLocation,
    BlowoutParams,
    BlowoutProperties,
)

from . import (
    boolean_looking_values,
    invalid_values,
    negative_or_zero_floats_and_ints,
    positive_non_zero_floats_and_ints,
)
from opentrons.protocol_api._liquid_properties import (
    _build_blowout_properties,
)


def test_blowout_properties_enable_and_disable() -> None:
    """Test that enable and disable work as expected."""
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=False,
            params=BlowoutParams(location=BlowoutLocation.DESTINATION, flowRate=100),
        )
    )
    bp.enabled = True
    assert bp.enabled is True
    bp.enabled = False
    assert bp.enabled is False


def test_blowout_properties_none_instantiation_combos() -> None:
    """Test that none values raise."""
    with pytest.raises(ValidationError):
        _build_blowout_properties(
            BlowoutProperties(
                enable=None,  # type: ignore
                params=BlowoutParams(location=None, flowRate=None),  # type: ignore
            )
        )
    with pytest.raises(ValidationError):
        _build_blowout_properties(
            BlowoutProperties(
                enable=True,
                params=BlowoutParams(location=None, flowRate=100),  # type: ignore
            )
        )


@given(bad_enable=boolean_looking_values)
@settings(deadline=None, max_examples=50)
def test_blowout_properties_enabled_bad_values(bad_enable: Any) -> None:
    """Test that invalid enable values raise."""
    with pytest.raises(ValidationError):
        _build_blowout_properties(
            BlowoutProperties(
                enable=bad_enable,
                params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=50),
            )
        )
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=False,
            params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=50),
        )
    )
    with pytest.raises(ValueError):
        bp.enabled = bad_enable


@given(good_flow_rate=positive_non_zero_floats_and_ints)
@settings(deadline=None, max_examples=50)
def test_blowout_properties_flow_rate(good_flow_rate: Any) -> None:
    """Test that valid flow rate values are accepted."""
    _build_blowout_properties(
        BlowoutProperties(
            enable=True,
            params=BlowoutParams(
                location=BlowoutLocation.DESTINATION, flowRate=good_flow_rate
            ),
        )
    )
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True,
            params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=1),
        )
    )
    bp.flow_rate = good_flow_rate
    assert bp.flow_rate == float(good_flow_rate)


@given(bad_flow_rate=st.one_of(invalid_values, negative_or_zero_floats_and_ints))
@settings(deadline=None, max_examples=50)
def test_blowout_properties_flow_rate_bad_values(bad_flow_rate: Any) -> None:
    """Test that invalid flow rate values raise."""
    with pytest.raises(ValidationError):
        _build_blowout_properties(
            BlowoutProperties(
                enable=True,
                params=BlowoutParams(
                    location=BlowoutLocation.TRASH, flowRate=bad_flow_rate
                ),
            )
        )
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True,
            params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=50),
        )
    )
    with pytest.raises(ValueError):
        bp.flow_rate = bad_flow_rate


@given(
    good_location=st.one_of(
        st.just(BlowoutLocation.DESTINATION),
        st.just(BlowoutLocation.TRASH),
        st.just(BlowoutLocation.SOURCE),
    )
)
@settings(deadline=None, max_examples=50)
def test_blowout_properties_location_enum(good_location: Any) -> None:
    """Test that valid location values are accepted."""
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True, params=BlowoutParams(location=good_location, flowRate=50)
        )
    )
    assert bp.location == good_location
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True,
            params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=50),
        )
    )

    bp.location = good_location
    assert bp.location == good_location


@given(
    good_location=st.one_of(
        st.just("destination"),
        st.just("trash"),
        st.just("source"),
    )
)
@settings(deadline=None, max_examples=50)
def test_blowout_properties_location_str(good_location: Any) -> None:
    """Test that valid location values are accepted."""
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True, params=BlowoutParams(location=good_location, flowRate=50)
        )
    )
    assert bp.location is not None and bp.location.value == good_location
    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True,
            params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=50),
        )
    )
    bp.location = good_location
    assert bp.location is not None and bp.location.value == good_location


@given(bad_location=st.one_of(invalid_values, st.just("chute")))
@settings(deadline=None, max_examples=50)
def test_blowout_properties_location_bad(bad_location: Any) -> None:
    """Test that invalid location values raise."""
    with pytest.raises(ValidationError):
        bp = _build_blowout_properties(
            BlowoutProperties(
                enable=True, params=BlowoutParams(location=bad_location, flowRate=50)
            )
        )

    bp = _build_blowout_properties(
        BlowoutProperties(
            enable=True,
            params=BlowoutParams(location=BlowoutLocation.TRASH, flowRate=50),
        )
    )
    with pytest.raises(ValueError):
        bp.location = bad_location



"""Tests for delay properties in the Opentrons protocol API."""

from typing import Any, Union

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st
from pydantic import ValidationError

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    DelayParams,
    DelayProperties,
)

from . import (
    boolean_looking_values,
    invalid_values,
    negative_non_zero_floats_and_ints,
    positive_or_zero_floats_or_ints,
)
from opentrons.protocol_api._liquid_properties import _build_delay_properties


def test_delay_properties_enable_and_disable() -> None:
    """Test enabling and disabling for boolean-only delay properties."""
    dp = _build_delay_properties(
        DelayProperties(enable=False, params=DelayParams(duration=1))
    )
    dp.enabled = True
    assert dp.enabled is True
    dp.enabled = False
    assert dp.enabled is False


def test_delay_properties_none_instantiation_combos() -> None:
    """Test handling of None combinations in delay properties instantiation."""
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=None, params=DelayParams(duration=None))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=False, params=DelayParams(duration=None))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=None, params=DelayParams(duration=1))  # type: ignore
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=True, params=DelayParams(duration=None))  # type: ignore
        )


@given(bad_value=st.one_of(invalid_values, boolean_looking_values))
@settings(deadline=None, max_examples=50)
def test_delay_properties_enabled_bad_values(bad_value: Any) -> None:
    """Test bad values for DelayProperties.enabled."""
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=bad_value, params=DelayParams(duration=1))
        )
    dp = _build_delay_properties(
        DelayProperties(enable=True, params=DelayParams(duration=1))
    )
    with pytest.raises(ValueError):
        dp.enabled = bad_value


@given(good_duration=positive_or_zero_floats_or_ints)
@settings(deadline=None, max_examples=50)
def test_delay_properties_duration(good_duration: Union[int, float]) -> None:
    """Test valid float/int >= 0 for DelayProperties."""
    _build_delay_properties(
        DelayProperties(enable=False, params=DelayParams(duration=good_duration))
    )
    dp = _build_delay_properties(
        DelayProperties(enable=True, params=DelayParams(duration=1))
    )
    dp.duration = good_duration
    assert dp.duration == float(good_duration)


@given(bad_duration=st.one_of(negative_non_zero_floats_and_ints, invalid_values))
@settings(deadline=None, max_examples=50)
def test_delay_properties_duration_bad_values(bad_duration: Any) -> None:
    """Test invalid float/int for DelayProperties (must be >= 0)."""
    # instantiation
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=True, params=DelayParams(duration=bad_duration))
        )
    with pytest.raises(ValidationError):
        _build_delay_properties(
            DelayProperties(enable=False, params=DelayParams(duration=bad_duration))
        )
    # setting
    dp = _build_delay_properties(
        DelayProperties(enable=True, params=DelayParams(duration=1))
    )
    with pytest.raises(ValueError):
        dp.duration = bad_duration
    dp = _build_delay_properties(
        DelayProperties(enable=False, params=DelayParams(duration=1))
    )
    with pytest.raises(ValueError):
        dp.duration = bad_duration



"""Unit tests for LabwareOffsetProvider."""

from datetime import datetime

import pytest
from decoy import Decoy

from opentrons.hardware_control.modules.types import TemperatureModuleModel
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocol_api.core.legacy.labware_offset_provider import (
    LabwareOffsetProvider,
    ProvidedLabwareOffset,
)
from opentrons.protocol_engine import (
    LabwareOffset,
    LabwareOffsetVector,
    LegacyLabwareOffsetLocation,
    ModuleModel,
    ProtocolEngine,
)
from opentrons.protocol_engine.state.labware import LabwareView
from opentrons.types import DeckSlotName, Point


@pytest.fixture
def protocol_engine(decoy: Decoy) -> ProtocolEngine:
    """Return a mock ProtocolEngine."""
    return decoy.mock(cls=ProtocolEngine)


@pytest.fixture
def labware_view(protocol_engine: ProtocolEngine) -> LabwareView:
    """Return a mock LabwareView."""
    return protocol_engine.state_view.labware


@pytest.fixture
def subject(protocol_engine: ProtocolEngine) -> LabwareOffsetProvider:
    """Return a LabwareOffsetProvider depending on the mocked LabwareView."""
    return LabwareOffsetProvider(engine=protocol_engine)


def test_find_something(
    subject: LabwareOffsetProvider, labware_view: LabwareView, decoy: Decoy
) -> None:
    """It should pass along simplified labware offset info from Protocol Engine."""
    decoy.when(
        labware_view.find_applicable_labware_offset_by_legacy_location(
            definition_uri="some_namespace/some_load_name/123",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V1,
            ),
        )
    ).then_return(
        LabwareOffset(
            # Subject should pass these along:
            id="labware-offset-id",
            vector=LabwareOffsetVector(x=1, y=2, z=3),
            # Shouldn't matter; subject should throw these away:
            definitionUri="result_definition_uri_should_not_matter",
            location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_11),
            createdAt=datetime(year=2021, month=1, day=1),
        )
    )

    result = subject.find(
        load_params=LabwareLoadParams("some_namespace", "some_load_name", 123),
        requested_module_model=TemperatureModuleModel.TEMPERATURE_V1,
        deck_slot=DeckSlotName.SLOT_1,
    )

    assert result == ProvidedLabwareOffset(
        delta=Point(x=1, y=2, z=3),
        offset_id="labware-offset-id",
    )


def test_find_nothing(
    subject: LabwareOffsetProvider, labware_view: LabwareView, decoy: Decoy
) -> None:
    """It should return a zero offset when Protocol Engine has no offset to provide."""
    decoy_call_rehearsal = (
        labware_view.find_applicable_labware_offset_by_legacy_location(
            definition_uri="some_namespace/some_load_name/123",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V1,
            ),
        )
    )

    decoy.when(decoy_call_rehearsal).then_return(None)

    result = subject.find(
        load_params=LabwareLoadParams("some_namespace", "some_load_name", 123),
        requested_module_model=TemperatureModuleModel.TEMPERATURE_V1,
        deck_slot=DeckSlotName.SLOT_1,
    )

    assert result == ProvidedLabwareOffset(delta=Point(x=0, y=0, z=0), offset_id=None)



from typing import List, Optional, Union

import pytest

from opentrons.protocols.parameters import validation as subject
from opentrons.protocols.parameters.exceptions import (
    ParameterDefinitionError,
    ParameterNameError,
    ParameterValueError,
)
from opentrons.protocols.parameters.types import (
    ParameterChoice,
    PrimitiveAllowedTypes,
)


def test_validate_variable_name_unique() -> None:
    """It should no-op if the name is unique or if it's not a string, and raise if it is not."""
    subject.validate_variable_name_unique("one of a kind", {"fee", "foo", "fum"})
    subject.validate_variable_name_unique({}, {"fee", "foo", "fum"})  # type: ignore[arg-type]
    with pytest.raises(ParameterNameError):
        subject.validate_variable_name_unique("copy", {"paste", "copy", "cut"})


def test_ensure_display_name() -> None:
    """It should ensure the display name is within the character limit."""
    result = subject.ensure_display_name("abc")
    assert result == "abc"


def test_ensure_display_name_raises() -> None:
    """It should raise if the display name is too long."""
    with pytest.raises(ParameterNameError):
        subject.ensure_display_name("Lorem ipsum dolor sit amet nam.")


def test_ensure_description_name() -> None:
    """It should ensure the description name is within the character limit."""
    result = subject.ensure_description("123456789")
    assert result == "123456789"


def test_ensure_description_raises() -> None:
    """It should raise if the description is too long."""
    with pytest.raises(ParameterNameError):
        subject.ensure_description(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
            " Fusce eget elementum nunc, quis sodales sed."
        )


def test_ensure_unit_string_length() -> None:
    """It should ensure the unit name is within the character limit."""
    result = subject.ensure_unit_string_length("ul")
    assert result == "ul"


def test_ensure_unit_string_length_raises() -> None:
    """It should raise if the unit name is too long."""
    with pytest.raises(ParameterNameError):
        subject.ensure_unit_string_length("newtons per square foot")


@pytest.mark.parametrize(
    "variable_name",
    [
        "x",
        "my_cool_variable",
        "_secret_variable",
    ],
)
def test_ensure_variable_name(variable_name: str) -> None:
    """It should ensure the variable name is a valid python variable name."""
    result = subject.ensure_variable_name(variable_name)
    assert result == variable_name


@pytest.mark.parametrize(
    "variable_name",
    [
        "3d_vector",
        "my cool variable name",
        "ca$h_money",
    ],
)
def test_ensure_variable_name_raises(variable_name: str) -> None:
    """It should raise if the variable name is not valid."""
    with pytest.raises(ParameterNameError, match="underscore"):
        subject.ensure_variable_name(variable_name)


@pytest.mark.parametrize(
    "variable_name",
    [
        "def",
        "class",
        "lambda",
    ],
)
def test_ensure_variable_name_raises_keyword(variable_name: str) -> None:
    """It should raise if the variable name is a python keyword."""
    with pytest.raises(ParameterNameError, match="keyword"):
        subject.ensure_variable_name(variable_name)


def test_validate_options() -> None:
    """It should not raise when given valid constraints"""
    subject.validate_options(123, 1, 100, None, int)
    subject.validate_options(123, 100, 100, None, int)
    subject.validate_options(
        123, None, None, [{"display_name": "abc", "value": 456}], int
    )
    subject.validate_options(12.3, 1.1, 100.9, None, float)
    subject.validate_options(12.3, 1.1, 1.1, None, float)
    subject.validate_options(
        12.3, None, None, [{"display_name": "abc", "value": 45.6}], float
    )
    subject.validate_options(
        True, None, None, [{"display_name": "abc", "value": False}], bool
    )
    subject.validate_options(
        "x", None, None, [{"display_name": "abc", "value": "y"}], str
    )


def test_validate_options_raises_value_error() -> None:
    """It should raise if the value of the default does not match the type."""
    with pytest.raises(ParameterValueError):
        subject.validate_options(123, 1, 100, None, str)


def test_validate_options_raises_name_error() -> None:
    """It should raise if the display name of a choice is too long."""
    with pytest.raises(ParameterNameError):
        subject.validate_options(
            "foo",
            None,
            None,
            [{"display_name": "Lorem ipsum dolor sit amet nam.", "value": "a"}],
            str,
        )


@pytest.mark.parametrize(
    ["value", "param_type", "result"],
    [
        (1.0, int, 1),
        (1.1, int, 1.1),
        (2, float, 2.0),
        (2.0, float, 2.0),
        (2.2, float, 2.2),
        ("3.0", str, "3.0"),
        (0.0, bool, False),
        (1, bool, True),
        (3.0, bool, 3.0),
        (True, bool, True),
    ],
)
def test_ensure_value_type(
    value: Union[float, bool, str], param_type: type, result: PrimitiveAllowedTypes
) -> None:
    """It should ensure that if applicable, the value is coerced into the expected type"""
    assert result == subject.ensure_value_type(value, param_type)


@pytest.mark.parametrize(
    ["value", "result"],
    [
        (1, 1.0),
        (2.0, 2.0),
        (3.3, 3.3),
    ],
)
def test_ensure_float_value(value: Union[float, int], result: float) -> None:
    """It should ensure that if applicable, the value is coerced into a float."""
    assert result == subject.ensure_float_value(value)


@pytest.mark.parametrize(
    ["value", "result"],
    [
        (1, 1.0),
        (2.0, 2.0),
        (3.3, 3.3),
        (None, None),
    ],
)
def test_ensure_optional_float_value(value: Union[float, int], result: float) -> None:
    """It should ensure that if applicable, the value is coerced into a float."""
    assert result == subject.ensure_optional_float_value(value)


@pytest.mark.parametrize(
    ["choices", "result"],
    [
        ([], []),
        (None, None),
        (
            [{"display_name": "foo", "value": 1}],
            [{"display_name": "foo", "value": 1.0}],
        ),
        (
            [{"display_name": "foo", "value": 2.0}],
            [{"display_name": "foo", "value": 2.0}],
        ),
        (
            [{"display_name": "foo", "value": 3.3}],
            [{"display_name": "foo", "value": 3.3}],
        ),
        (
            [{"display_name": "foo", "value": "4"}],
            [{"display_name": "foo", "value": "4"}],
        ),
        (
            [{"display_name": "foo", "value": True}],
            [{"display_name": "foo", "value": True}],
        ),
    ],
)
def test_ensure_float_choices(
    choices: Optional[List[ParameterChoice]], result: Optional[List[ParameterChoice]]
) -> None:
    """It should ensure that if applicable, the value in a choice is coerced into a float."""
    assert result == subject.ensure_float_choices(choices)


@pytest.mark.parametrize(
    ["param_type", "result"],
    [(int, "int"), (float, "float"), (str, "str")],
)
def test_convert_type_string_for_enum(param_type: type, result: str) -> None:
    """It should convert the type into a string for the EnumParameter model."""
    assert result == subject.convert_type_string_for_enum(param_type)


def test_convert_type_string_for_enum_raises() -> None:
    """It should raise if given a bool to convert to an enum type string."""
    with pytest.raises(ParameterValueError):
        subject.convert_type_string_for_enum(bool)


@pytest.mark.parametrize(["param_type", "result"], [(int, "int"), (float, "float")])
def test_convert_type_string_for_num_param(param_type: type, result: str) -> None:
    """It should convert the type into a string for the NumberParameter model."""
    assert result == subject.convert_type_string_for_num_param(param_type)


@pytest.mark.parametrize(
    "param_type",
    [str, bool],
)
def test_convert_type_string_for_num_param_raises(param_type: type) -> None:
    """It should raise if given a bool or str to convert to a number type string."""
    with pytest.raises(ParameterValueError):
        subject.convert_type_string_for_num_param(param_type)


@pytest.mark.parametrize(
    ["default", "minimum", "maximum", "choices", "parameter_type", "error_text"],
    [
        (123, None, None, None, int, "provide either"),
        (
            123,
            1,
            None,
            [{"display_name": "abc", "value": 123}],
            int,
            "maximum values cannot",
        ),
        (
            123,
            None,
            100,
            [{"display_name": "abc", "value": 123}],
            int,
            "maximum values cannot",
        ),
        (123, None, None, [{"display_name": "abc"}], int, "dictionary with keys"),
        (123, None, None, [{"value": 123}], int, "dictionary with keys"),
        (
            123,
            None,
            None,
            [{"display_name": "abc", "value": "123"}],
            int,
            "must be of type",
        ),
        (123, 1, None, None, int, "maximum must also"),
        (123, None, 100, None, int, "minimum must also"),
        (123, 100, 1, None, int, "Maximum must be greater"),
        (123, 1.1, 100, None, int, "Minimum is type"),
        (123, 1, 100.5, None, int, "Maximum is type"),
        (123.0, "1.0", 100.0, None, float, "Minimum is type"),
        ("blah", 1, 100, None, str, "Only parameters of type float or int"),
    ],
)
def test_validate_options_raise_definition_error(
    default: PrimitiveAllowedTypes,
    minimum: Optional[PrimitiveAllowedTypes],
    maximum: Optional[PrimitiveAllowedTypes],
    choices: Optional[List[ParameterChoice]],
    parameter_type: type,
    error_text: str,
) -> None:
    """It should raise if the parameter definition constraints are not valid."""
    with pytest.raises(ParameterDefinitionError, match=error_text):
        subject.validate_options(default, minimum, maximum, choices, parameter_type)



"""Test Flex Stacker retrieve command implementation."""

from datetime import datetime
from typing import Type, Union
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    FlexStackerHopperLabwareError,
    FlexStackerShuttleLabwareError,
    FlexStackerShuttleMissingError,
    FlexStackerShuttleNotEmptyError,
    FlexStackerStallError,
)
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
)

from opentrons.drivers.flex_stacker.types import StackerAxis
from opentrons.hardware_control.modules import FlexStacker, PlatformState
from opentrons.protocol_engine.commands import flex_stacker
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.flex_stacker.common import (
    FlexStackerHopperError,
    FlexStackerLabwareRetrieveError,
    FlexStackerShuttleError,
    FlexStackerShuttleOccupiedError,
    FlexStackerStallOrCollisionError,
)
from opentrons.protocol_engine.commands.flex_stacker.retrieve import RetrieveImpl
from opentrons.protocol_engine.errors import CannotPerformModuleAction
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    AddressableAreaUsedUpdate,
    BatchLabwareLocationUpdate,
    FlexStackerStateUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    InStackerHopperLocation,
    LabwareLocationSequence,
    LabwareOffset,
    LabwareUri,
    LoadedModule,
    ModuleLocation,
    ModuleModel,
    OnAddressableAreaLocationSequenceComponent,
    OnCutoutFixtureLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
    OnLabwareOffsetLocationSequenceComponent,
    OnModuleLocationSequenceComponent,
    StackerStoredLabwareGroup,
)
from opentrons.types import DeckSlotName


def _contained_labware(
    count: int, with_lid: bool = False, with_adapter: bool = False
) -> list[StackerStoredLabwareGroup]:
    return [
        StackerStoredLabwareGroup(
            primaryLabwareId=f"primary-id-{i + 1}",
            adapterLabwareId=f"adapter-id-{i + 1}" if with_adapter else None,
            lidLabwareId=f"lid-id-{i + 1}" if with_lid else None,
        )
        for i in range(count)
    ]


def _prep_stacker_own_location(
    decoy: Decoy, state_view: StateView, stacker_id: str
) -> None:
    decoy.when(state_view.modules.get_location(stacker_id)).then_return(
        DeckSlotLocation(slotName=DeckSlotName(value="B3")),
    )
    decoy.when(state_view.modules.get(stacker_id)).then_return(
        LoadedModule(
            id=stacker_id,
            location=DeckSlotLocation(slotName=DeckSlotName(value="B3")),
            model=ModuleModel.FLEX_STACKER_MODULE_V1,
            serialNumber="HIIIII",
        )
    )
    decoy.when(
        state_view.modules.ensure_and_convert_module_fixture_location(
            deck_slot=DeckSlotName("B3"), model=ModuleModel.FLEX_STACKER_MODULE_V1
        )
    ).then_return("flexStackerV1B4")


def _stacker_base_loc_seq(stacker_id: str) -> LabwareLocationSequence:
    return [
        OnAddressableAreaLocationSequenceComponent(
            addressableAreaName="flexStackerV1B4"
        ),
        OnModuleLocationSequenceComponent(moduleId=stacker_id),
        OnCutoutFixtureLocationSequenceComponent(
            cutoutId="cutoutB3", possibleCutoutFixtureIds=["flexStackerModuleV1"]
        ),
    ]


@pytest.fixture
def subject(
    state_view: StateView, equipment: EquipmentHandler, model_utils: ModelUtils
) -> RetrieveImpl:
    """Get a retrieve command to test."""
    return RetrieveImpl(
        state_view=state_view, equipment=equipment, model_utils=model_utils
    )


async def test_retrieve_raises_when_empty(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: RetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_id: FlexStackerId,
) -> None:
    """It should raise an exception when called on an empty pool."""
    data = flex_stacker.RetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    with pytest.raises(
        CannotPerformModuleAction,
        match="Cannot retrieve labware from Flex Stacker in .* because it contains no labware",
    ):
        await subject.execute(data)


async def test_retrieve_primary_only(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: RetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware."""
    data = flex_stacker.RetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(1),
        max_pool_count=5,
        pool_overlap=6,
        pool_height=10,
    )
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_projected_offset_location(
            ModuleLocation(moduleId=stacker_id)
        )
    ).then_return(sentinel.primary_offset_location)
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            sentinel.primary_offset_location,
        )
    ).then_return(
        LabwareOffset.model_construct(id="offset-id-1")  # type: ignore[call-arg]
    )

    _prep_stacker_own_location(decoy, state_view, stacker_id)

    result = await subject.execute(data)

    decoy.verify(
        await stacker_hardware.dispense_labware(labware_height=4),
        times=1,
    )

    assert result == SuccessData(
        public=flex_stacker.RetrieveResult(
            labwareId="primary-id-1",
            primaryLocationSequence=_stacker_base_loc_seq(stacker_id),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            originalPrimaryLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "primary-id-1": ModuleLocation(moduleId=stacker_id)
                },
                new_offset_ids_by_id={"primary-id-1": "offset-id-1"},
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id, contained_labware_bottom_first=[]
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_retrieve_primary_and_lid(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: RetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_lid_def: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware with a lid on it."""
    data = flex_stacker.RetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=tiprack_lid_def,
        contained_labware_bottom_first=_contained_labware(2, with_lid=True),
        max_pool_count=5,
        pool_overlap=2,
        pool_height=10,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_lid_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_tiprack_lid/1"))
    decoy.when(state_view.labware.get_uri_from_definition(tiprack_lid_def)).then_return(
        LabwareUri("opentrons/opentrons_flex_tiprack_lid/1")
    )

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_projected_offset_location(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return([sentinel.primary_offset_location])
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            [sentinel.primary_offset_location],
        )
    ).then_return(
        LabwareOffset.model_construct(id="offset-id-1")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_tiprack_lid/1",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons/opentrons_flex_96_filtertiprack_50ul/1"
                ),
                sentinel.primary_offset_location,
            ],
        )
    ).then_return(None)
    _prep_stacker_own_location(decoy, state_view, stacker_id)

    result = await subject.execute(data)

    decoy.verify(
        await stacker_hardware.dispense_labware(labware_height=8),
        times=1,
    )

    assert result == SuccessData(
        public=flex_stacker.RetrieveResult(
            labwareId="primary-id-1",
            lidId="lid-id-1",
            primaryLocationSequence=_stacker_base_loc_seq(stacker_id),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            lidLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="primary-id-1", lidId="lid-id-1"
                    )
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            originalPrimaryLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
            originalLidLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="primary-id-1", lidId="lid-id-1"
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            lidLabwareURI="opentrons/opentrons_flex_tiprack_lid/1",
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "primary-id-1": ModuleLocation(moduleId=stacker_id),
                    "lid-id-1": OnLabwareLocation(labwareId="primary-id-1"),
                },
                new_offset_ids_by_id={"primary-id-1": "offset-id-1", "lid-id-1": None},
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[
                    _contained_labware(2, with_lid=True)[1]
                ],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_retrieve_primary_and_adapter(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: RetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_adapter_def: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware on an adapter."""
    data = flex_stacker.RetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=tiprack_adapter_def,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(2, with_adapter=True),
        max_pool_count=5,
        pool_overlap=3,
        pool_height=15,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_adapter_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_tiprack_adapter/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition(tiprack_adapter_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_tiprack_adapter/1"))

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id)
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_projected_offset_location(
            ModuleLocation(moduleId=stacker_id)
        )
    ).then_return([sentinel.adapter_offset_location])
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_96_tiprack_adapter/1",
            [sentinel.adapter_offset_location],
        )
    ).then_return(
        LabwareOffset.model_construct(id="offset-id-1")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons/opentrons_flex_96_tiprack_adapter/1"
                ),
                sentinel.adapter_offset_location,
            ],
        )
    ).then_return(
        LabwareOffset.model_construct(id="offset-id-2")  # type: ignore[call-arg]
    )

    _prep_stacker_own_location(decoy, state_view, stacker_id)
    result = await subject.execute(data)

    decoy.verify(
        await stacker_hardware.dispense_labware(labware_height=12),
        times=1,
    )

    assert result == SuccessData(
        public=flex_stacker.RetrieveResult(
            labwareId="primary-id-1",
            adapterId="adapter-id-1",
            primaryLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    )
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            adapterLocationSequence=_stacker_base_loc_seq(stacker_id),
            adapterLabwareURI="opentrons/opentrons_flex_96_tiprack_adapter/1",
            originalPrimaryLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="adapter-id-1", lidId=None
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            originalAdapterLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "adapter-id-1": ModuleLocation(moduleId=stacker_id),
                    "primary-id-1": OnLabwareLocation(labwareId="adapter-id-1"),
                },
                new_offset_ids_by_id={
                    "primary-id-1": "offset-id-2",
                    "adapter-id-1": "offset-id-1",
                },
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[
                    _contained_labware(2, with_adapter=True)[1]
                ],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_retrieve_primary_adapter_and_lid(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: RetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_adapter_def: LabwareDefinition,
    tiprack_lid_def: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware on an adapter."""
    data = flex_stacker.RetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=tiprack_adapter_def,
        pool_lid_definition=tiprack_lid_def,
        contained_labware_bottom_first=_contained_labware(
            2, with_lid=True, with_adapter=True
        ),
        max_pool_count=5,
        pool_overlap=4,
        pool_height=20,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_lid_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_tiprack_lid/1"))
    decoy.when(state_view.labware.get_uri_from_definition(tiprack_lid_def)).then_return(
        LabwareUri("opentrons/opentrons_flex_tiprack_lid/1")
    )
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_adapter_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_tiprack_adapter/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition(tiprack_adapter_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_tiprack_adapter/1"))

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_projected_offset_location(
            ModuleLocation(moduleId=stacker_id)
        )
    ).then_return([sentinel.adapter_offset_location])
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_96_tiprack_adapter/1",
            [sentinel.adapter_offset_location],
        )
    ).then_return(
        LabwareOffset.model_construct(id="offset-id-1")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons/opentrons_flex_96_tiprack_adapter/1"
                ),
                sentinel.adapter_offset_location,
            ],
        )
    ).then_return(
        LabwareOffset.model_construct(id="offset-id-2")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.labware.find_applicable_labware_offset(
            "opentrons/opentrons_flex_tiprack_lid/1",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons/opentrons_flex_96_filtertiprack_50ul/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons/opentrons_flex_96_tiprack_adapter/1"
                ),
                sentinel.adapter_offset_location,
            ],
        )
    ).then_return(None)

    _prep_stacker_own_location(decoy, state_view, stacker_id)
    result = await subject.execute(data)

    decoy.verify(
        await stacker_hardware.dispense_labware(
            labware_height=16,
        ),
        times=1,
    )

    assert result == SuccessData(
        public=flex_stacker.RetrieveResult(
            labwareId="primary-id-1",
            adapterId="adapter-id-1",
            lidId="lid-id-1",
            primaryLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    )
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            adapterLocationSequence=_stacker_base_loc_seq(stacker_id),
            adapterLabwareURI="opentrons/opentrons_flex_96_tiprack_adapter/1",
            lidLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="primary-id-1", lidId="lid-id-1"
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    ),
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            lidLabwareURI="opentrons/opentrons_flex_tiprack_lid/1",
            originalPrimaryLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="adapter-id-1", lidId=None
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            originalLidLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="primary-id-1", lidId="lid-id-1"
                ),
                OnLabwareLocationSequenceComponent(
                    labwareId="adapter-id-1", lidId=None
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            originalAdapterLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "adapter-id-1": ModuleLocation(moduleId=stacker_id),
                    "primary-id-1": OnLabwareLocation(labwareId="adapter-id-1"),
                    "lid-id-1": OnLabwareLocation(labwareId="primary-id-1"),
                },
                new_offset_ids_by_id={
                    "primary-id-1": "offset-id-2",
                    "adapter-id-1": "offset-id-1",
                    "lid-id-1": None,
                },
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[
                    _contained_labware(2, with_adapter=True, with_lid=True)[1]
                ],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


@pytest.mark.parametrize(
    "shared_data_error,protocol_engine_error",
    [
        (
            FlexStackerStallError(serial="123", axis=StackerAxis.Z),
            FlexStackerStallOrCollisionError,
        ),
        (
            FlexStackerShuttleMissingError(
                serial="123",
                expected_state=PlatformState.EXTENDED,
                shuttle_state=PlatformState.UNKNOWN,
            ),
            FlexStackerShuttleError,
        ),
        (
            FlexStackerHopperLabwareError(
                serial="123",
                labware_expected=True,
            ),
            FlexStackerHopperError,
        ),
        (
            FlexStackerShuttleLabwareError(
                serial="123", labware_expected=True, shuttle_state=""
            ),
            FlexStackerLabwareRetrieveError,
        ),
        (
            FlexStackerShuttleNotEmptyError(
                serial="123", labware_expected=True, shuttle_state=""
            ),
            FlexStackerShuttleOccupiedError,
        ),
    ],
)
async def test_retrieve_raises_recoverable_error(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    subject: RetrieveImpl,
    model_utils: ModelUtils,
    stacker_id: FlexStackerId,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_hardware: FlexStacker,
    shared_data_error: Exception,
    protocol_engine_error: Type[
        Union[
            FlexStackerStallOrCollisionError,
            FlexStackerShuttleError,
            FlexStackerLabwareRetrieveError,
        ]
    ],
) -> None:
    """It should raise a stall error."""
    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    data = flex_stacker.RetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(1),
        max_pool_count=999,
        pool_overlap=4,
        pool_height=20,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    _prep_stacker_own_location(decoy, state_view, stacker_id)

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    decoy.when(
        await stacker_hardware.dispense_labware(  # type: ignore[func-returns-value]
            labware_height=16,
        )
    ).then_raise(shared_data_error)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=protocol_engine_error.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={"labwareId": "primary-id-1"},
        ),
        state_update_if_false_positive=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "primary-id-1": ModuleLocation(moduleId=stacker_id),
                },
                new_offset_ids_by_id={
                    "primary-id-1": None,
                },
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )



"""Tests for mix properties in the Opentrons protocol API."""

from typing import Any, Union

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st
from pydantic import ValidationError

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    MixParams,
    MixProperties,
)

from . import (
    boolean_looking_values,
    invalid_values,
    negative_or_zero_floats_and_ints,
    positive_non_zero_floats_and_ints,
)
from opentrons.protocol_api._liquid_properties import _build_mix_properties


def test_mix_properties_enable_and_disable() -> None:
    """Test enabling and disabling for boolean-only mix properties."""
    mp = _build_mix_properties(
        MixProperties(enable=False, params=MixParams(repetitions=2, volume=10))
    )
    mp.enabled = True
    assert mp.enabled is True
    mp.enabled = False
    assert mp.enabled is False


def test_mix_properties_none_instantiation_combos() -> None:
    """Test handling of None combos in mix properties instantiation."""
    with pytest.raises(ValidationError):
        _build_mix_properties(MixProperties(enable=True, params=None))
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(enable=None, params=MixParams(repetitions=2, volume=10))  # type: ignore
        )
    _build_mix_properties(MixProperties(enable=False, params=None))


@given(bad_value=st.one_of(invalid_values, boolean_looking_values))
@settings(deadline=None, max_examples=50)
def test_mix_properties_enabled_bad_values(bad_value: Any) -> None:
    """Test bad values for MixProperties.enabled."""
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(enable=bad_value, params=MixParams(repetitions=1, volume=5))
        )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=10))
    )
    with pytest.raises(ValueError):
        mp.enabled = bad_value


@given(good_volume=positive_non_zero_floats_and_ints)
@settings(deadline=None, max_examples=50)
def test_mix_properties_volume_good_values(good_volume: Union[int, float]) -> None:
    """Test valid float/int > 0 for MixProperties volume."""
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=5))
    )
    mp.volume = good_volume
    assert mp.volume == float(good_volume)


@given(bad_volume=st.one_of(negative_or_zero_floats_and_ints, invalid_values))
@settings(deadline=None, max_examples=50)
def test_mix_properties_volume_bad_values(bad_volume: Any) -> None:
    """Test invalid float/int <= 0 for MixProperties volume."""
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(
                enable=True, params=MixParams(repetitions=2, volume=bad_volume)
            )
        )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=5))
    )
    with pytest.raises(ValueError):
        mp.volume = bad_volume


@given(good_reps=st.integers(min_value=0, max_value=100))
@settings(deadline=None, max_examples=50)
def test_mix_properties_repetitions_good_values(good_reps: int) -> None:
    """Test valid int >= 0 for MixProperties repetitions."""
    _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=good_reps, volume=5))
    )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=2, volume=5))
    )
    mp.repetitions = good_reps
    assert mp.repetitions == good_reps


@given(bad_reps=st.one_of(st.integers(max_value=-1), invalid_values))
@settings(deadline=None, max_examples=50)
def test_mix_properties_repetitions_bad_values(bad_reps: Any) -> None:
    """Test invalid repetitions < 1 or non-integer."""
    with pytest.raises(ValidationError):
        _build_mix_properties(
            MixProperties(enable=True, params=MixParams(repetitions=bad_reps, volume=5))
        )
    mp = _build_mix_properties(
        MixProperties(enable=True, params=MixParams(repetitions=3, volume=5))
    )
    with pytest.raises(ValueError):
        mp.repetitions = bad_reps



"""Tests for the ProtocolReader interface."""

from __future__ import annotations

from pathlib import Path
from typing import Optional

import pytest
from decoy import Decoy, matchers

from opentrons.protocol_reader import (
    FileHasher,
    ProtocolFileRole,
    ProtocolReader,
    ProtocolSource,
    ProtocolSourceFile,
    PythonProtocolConfig,
)
from opentrons.protocol_reader.file_format_validator import FileFormatValidator
from opentrons.protocol_reader.file_identifier import (
    FileIdentifier,
    IdentifiedData,
    IdentifiedLabwareDefinition,
    IdentifiedPythonMain,
)
from opentrons.protocol_reader.file_reader_writer import (
    BufferedFile,
    FileReaderWriter,
)
from opentrons.protocol_reader.role_analyzer import (
    RoleAnalysis,
    RoleAnalyzer,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parse import PythonParseMode


@pytest.fixture
def file_reader_writer(decoy: Decoy) -> FileReaderWriter:
    """Get a mocked out FileReaderWriter."""
    return decoy.mock(cls=FileReaderWriter)


@pytest.fixture
def file_identifier(decoy: Decoy) -> FileIdentifier:
    """Get a mocked out FileIdentifier."""
    return decoy.mock(cls=FileIdentifier)


@pytest.fixture
def role_analyzer(decoy: Decoy) -> RoleAnalyzer:
    """Get a mocked out RoleAnalyzer."""
    return decoy.mock(cls=RoleAnalyzer)


@pytest.fixture
def file_format_validator(decoy: Decoy) -> FileFormatValidator:
    """Get a mocked out FileFormatValidator."""
    return decoy.mock(cls=FileFormatValidator)


@pytest.fixture
def file_hasher(decoy: Decoy) -> FileHasher:
    """Get a mocked out FileHasher."""
    return decoy.mock(cls=FileHasher)


@pytest.fixture
def subject(
    file_reader_writer: FileReaderWriter,
    role_analyzer: RoleAnalyzer,
    file_identifier: FileIdentifier,
    file_format_validator: FileFormatValidator,
    file_hasher: FileHasher,
) -> ProtocolReader:
    """Create a ProtocolReader test subject."""
    return ProtocolReader(
        file_reader_writer=file_reader_writer,
        role_analyzer=role_analyzer,
        file_identifier=file_identifier,
        file_format_validator=file_format_validator,
        file_hasher=file_hasher,
    )


async def test_save(
    decoy: Decoy,
    tmp_path: Path,
    file_reader_writer: FileReaderWriter,
    file_identifier: FileIdentifier,
    role_analyzer: RoleAnalyzer,
    file_format_validator: FileFormatValidator,
    subject: ProtocolReader,
) -> None:
    """It should compute a single file protocol source."""
    buffered_main_file = BufferedFile(
        name="protocol.py",
        contents=b"# hello world",
        path=None,
    )
    buffered_labware_file = BufferedFile(
        name="labware.json",
        contents=b"wow",
        path=None,
    )
    buffered_data_file = BufferedFile(
        name="data.txt",
        contents=b"beep boop",
        path=None,
    )

    main_file = IdentifiedPythonMain(
        original_file=buffered_main_file,
        api_level=APIVersion(123, 456),
        robot_type="OT-2 Standard",
        metadata={"hey": "there"},
    )
    labware_file = IdentifiedLabwareDefinition(
        original_file=buffered_labware_file,
        unvalidated_json={},
    )
    data_file = IdentifiedData(original_file=buffered_data_file)

    role_analysis = RoleAnalysis(
        main_file=main_file, labware_files=[labware_file], data_files=[data_file]
    )

    decoy.when(
        await file_identifier.identify(
            [buffered_main_file, buffered_labware_file, buffered_data_file],
            python_parse_mode=PythonParseMode.NORMAL,
        )
    ).then_return([main_file, labware_file, data_file])
    decoy.when(role_analyzer.analyze([main_file, labware_file, data_file])).then_return(
        role_analysis
    )

    result = await subject.save(
        files=[buffered_main_file, buffered_labware_file, buffered_data_file],
        directory=tmp_path,
        content_hash="abc123",
    )

    assert result == ProtocolSource(
        directory=tmp_path,
        main_file=tmp_path / "protocol.py",
        files=[
            ProtocolSourceFile(
                path=tmp_path / "protocol.py",
                role=ProtocolFileRole.MAIN,
            ),
            ProtocolSourceFile(
                path=tmp_path / "labware.json",
                role=ProtocolFileRole.LABWARE,
            ),
            ProtocolSourceFile(
                path=tmp_path / "data.txt",
                role=ProtocolFileRole.DATA,
            ),
        ],
        metadata={"hey": "there"},
        robot_type="OT-2 Standard",
        config=PythonProtocolConfig(api_version=APIVersion(123, 456)),
        content_hash="abc123",
    )

    decoy.verify(
        await file_format_validator.validate([main_file, labware_file, data_file]),
        await file_reader_writer.write(
            directory=tmp_path,
            files=[buffered_main_file, buffered_labware_file, buffered_data_file],
        ),
    )


@pytest.mark.parametrize("directory", [None, Path("/some/dir")])
@pytest.mark.parametrize(
    "files_are_prevalidated, validator_expected_times_called", [(True, 0), (False, 1)]
)
async def test_read_saved(
    decoy: Decoy,
    directory: Optional[Path],
    files_are_prevalidated: bool,
    validator_expected_times_called: int,
    file_reader_writer: FileReaderWriter,
    file_identifier: FileIdentifier,
    role_analyzer: RoleAnalyzer,
    file_format_validator: FileFormatValidator,
    file_hasher: FileHasher,
    subject: ProtocolReader,
) -> None:
    """It should read a single file protocol source without copying elsewhere."""
    input_main_file = Path("/path/to/protocol.py")
    input_labware_file = Path("/path/to/labware.json")
    input_data_file = Path("/path/to/data.txt")

    buffered_main_file = BufferedFile(
        name="protocol.py",
        contents=b"# hello world",
        path=Path("/path/to/protocol.py"),
    )
    buffered_labware_file = BufferedFile(
        name="labware.json",
        contents=b"wow",
        path=Path("/path/to/labware.json"),
    )
    buffered_data_file = BufferedFile(
        name="data.txt",
        contents=b"beep boop",
        path=Path("/path/to/data.txt"),
    )

    main_file = IdentifiedPythonMain(
        original_file=buffered_main_file,
        api_level=APIVersion(123, 456),
        robot_type="OT-2 Standard",
        metadata={"hey": "there"},
    )
    labware_file = IdentifiedLabwareDefinition(
        original_file=buffered_labware_file,
        unvalidated_json={},
    )
    data_file = IdentifiedData(original_file=buffered_data_file)

    role_analysis = RoleAnalysis(
        main_file=main_file,
        labware_files=[labware_file],
        data_files=[data_file],
    )

    decoy.when(
        await file_reader_writer.read(
            [input_main_file, input_labware_file, input_data_file]
        )
    ).then_return([buffered_main_file, buffered_labware_file, buffered_data_file])
    decoy.when(
        await file_identifier.identify(
            [buffered_main_file, buffered_labware_file, buffered_data_file],
            python_parse_mode=PythonParseMode.NORMAL,
        )
    ).then_return([main_file, labware_file, data_file])
    decoy.when(role_analyzer.analyze([main_file, labware_file, data_file])).then_return(
        role_analysis
    )
    decoy.when(
        await file_hasher.hash(
            [buffered_main_file, buffered_labware_file, buffered_data_file]
        )
    ).then_return("abc123")

    result = await subject.read_saved(
        files=[input_main_file, input_labware_file, input_data_file],
        directory=directory,
        files_are_prevalidated=files_are_prevalidated,
    )

    assert result == ProtocolSource(
        directory=directory,
        main_file=Path("/path/to/protocol.py"),
        files=[
            ProtocolSourceFile(
                path=Path("/path/to/protocol.py"),
                role=ProtocolFileRole.MAIN,
            ),
            ProtocolSourceFile(
                path=Path("/path/to/labware.json"),
                role=ProtocolFileRole.LABWARE,
            ),
            ProtocolSourceFile(
                path=Path("/path/to/data.txt"),
                role=ProtocolFileRole.DATA,
            ),
        ],
        metadata={"hey": "there"},
        robot_type="OT-2 Standard",
        config=PythonProtocolConfig(api_version=APIVersion(123, 456)),
        content_hash="abc123",
    )

    decoy.verify(
        await file_reader_writer.write(
            directory=matchers.Anything(),
            files=matchers.Anything(),
        ),
        times=0,
    )

    decoy.verify(
        await file_format_validator.validate([main_file, labware_file, data_file]),
        times=validator_expected_times_called,
    )



"""Test prepare to aspirate commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import PipetteOverpressureError

from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.pipetting_common import OverpressureError
from opentrons.protocol_engine.commands.prepare_to_aspirate import (
    PrepareToAspirateImplementation,
    PrepareToAspirateParams,
    PrepareToAspirateResult,
)
from opentrons.protocol_engine.execution import (
    PipettingHandler,
)
from opentrons.protocol_engine.execution.gantry_mover import GantryMover
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.types import Point


@pytest.fixture
def subject(
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
    gantry_mover: GantryMover,
) -> PrepareToAspirateImplementation:
    """Get the implementation subject."""
    return PrepareToAspirateImplementation(
        pipetting=pipetting, model_utils=model_utils, gantry_mover=gantry_mover
    )


async def test_prepare_to_aspirate_implementation(
    decoy: Decoy,
    gantry_mover: GantryMover,
    subject: PrepareToAspirateImplementation,
    pipetting: PipettingHandler,
) -> None:
    """A PrepareToAspirate command should have an executing implementation."""
    data = PrepareToAspirateParams(pipetteId="some id")
    position = Point(x=1, y=2, z=3)
    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id="some id")).then_return(
        False
    )
    decoy.when(await pipetting.prepare_for_aspirate(pipette_id="some id")).then_return(  # type: ignore[func-returns-value]
        None
    )
    decoy.when(await gantry_mover.get_position("some id")).then_return(position)

    result = await subject.execute(data)
    assert result == SuccessData(
        public=PrepareToAspirateResult(),
        state_update=update_types.StateUpdate(
            pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                pipette_id="some id",
                clean_tip=False,
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="some id", ready_to_aspirate=True
            ),
        ),
    )


async def test_overpressure_error(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    subject: PrepareToAspirateImplementation,
    model_utils: ModelUtils,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"

    position = Point(x=1, y=2, z=3)

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    data = PrepareToAspirateParams(
        pipetteId=pipette_id,
    )
    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id="pipette-id")).then_return(
        False
    )

    decoy.when(
        await pipetting.prepare_for_aspirate(  # type: ignore[func-returns-value]
            pipette_id=pipette_id,
        ),
    ).then_raise(PipetteOverpressureError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)
    decoy.when(await gantry_mover.get_position(pipette_id)).then_return(position)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=OverpressureError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={"retryLocation": (position.x, position.y, position.z)},
        ),
        state_update=update_types.StateUpdate(
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="pipette-id"
            )
        ),
    )


async def test_prepare_noops_if_prepared(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    subject: PrepareToAspirateImplementation,
    model_utils: ModelUtils,
) -> None:
    """It should do nothing if the pipette does not need to be prepared."""
    data = PrepareToAspirateParams(pipetteId="some id")
    position = Point(x=1, y=2, z=3)
    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id="some id")).then_return(
        True
    )
    decoy.when(await gantry_mover.get_position("some id")).then_return(position)

    result = await subject.execute(data)
    decoy.verify(await pipetting.prepare_for_aspirate(pipette_id="some id"), times=0)
    assert result == SuccessData(
        public=PrepareToAspirateResult(),
        state_update=update_types.StateUpdate(),
    )



"""Test pipette data provider."""

from sys import maxsize
from typing import Dict

import pytest

from opentrons_shared_data.pipette import pipette_definition
from opentrons_shared_data.pipette import types as pip_types
from opentrons_shared_data.pipette.pipette_definition import (
    TIP_OVERLAP_VERSION_MAXIMUM,
    AvailableSensorDefinition,
    PipetteBoundingBoxOffsetDefinition,
)
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteModel,
    PipetteNameType,
)

from ..pipette_fixtures import get_default_nozzle_map
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocol_engine.errors.exceptions import InvalidLoadPipetteSpecsError
from opentrons.protocol_engine.resources import pipette_data_provider as subject
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
    VirtualPipetteDataProvider,
    get_latest_tip_overlap_before_version,
    validate_and_default_tip_overlap_version,
)
from opentrons.protocol_engine.types import FlowRates
from opentrons.types import Point


@pytest.fixture
def available_sensors() -> AvailableSensorDefinition:
    """Provide a list of sensors."""
    return AvailableSensorDefinition(sensors=["pressure", "capacitive", "environment"])


@pytest.fixture
def subject_instance() -> VirtualPipetteDataProvider:
    """Instance of a VirtualPipetteDataProvider for test."""
    return VirtualPipetteDataProvider()


def test_get_virtual_pipette_static_config(
    subject_instance: VirtualPipetteDataProvider,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should return config data given a pipette name."""
    result = subject_instance.get_virtual_pipette_static_config(
        PipetteNameType.P20_SINGLE_GEN2.value, "some-id", "v0"
    )

    assert result == LoadedStaticPipetteData(
        model="p20_single_v2.2",
        display_name="P20 Single-Channel GEN2",
        min_volume=1,
        max_volume=20.0,
        channels=1,
        nozzle_offset_z=10.45,
        home_position=172.15,
        flow_rates=FlowRates(
            default_aspirate={"2.0": 3.78, "2.6": 7.56},
            default_dispense={"2.0": 3.78, "2.6": 7.56},
            default_blow_out={"2.0": 3.78, "2.6": 7.56},
        ),
        tip_configuration_lookup_table=result.tip_configuration_lookup_table,
        nominal_tip_overlap={
            "default": 8.25,
            "opentrons/eppendorf_96_tiprack_10ul_eptips/1": 8.4,
            "opentrons/geb_96_tiprack_10ul/1": 8.3,
            "opentrons/opentrons_96_filtertiprack_10ul/1": 8.25,
            "opentrons/opentrons_96_filtertiprack_20ul/1": 8.25,
            "opentrons/opentrons_96_tiprack_10ul/1": 8.25,
            "opentrons/opentrons_96_tiprack_20ul/1": 8.25,
        },
        nozzle_map=result.nozzle_map,
        back_left_corner_offset=Point(-16, 22.25, 10.45),
        front_right_corner_offset=Point(16, -22.25, 10.45),
        pipette_lld_settings={},
        plunger_positions={
            "top": 19.5,
            "bottom": -8.5,
            "blow_out": -13.0,
            "drop_tip": -27.0,
        },
        shaft_ul_per_mm=0.785,
        available_sensors=AvailableSensorDefinition(sensors=[]),
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={VolumeModes.default: 1.0},
    )


def test_configure_virtual_pipette_for_volume(
    subject_instance: VirtualPipetteDataProvider,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should return an updated config if the liquid class changes."""
    result1 = subject_instance.get_virtual_pipette_static_config(
        PipetteNameType.P50_SINGLE_FLEX.value, "my-pipette", "v0"
    )
    assert result1 == LoadedStaticPipetteData(
        model="p50_single_v3.6",
        display_name="Flex 1-Channel 50 µL",
        min_volume=5,
        max_volume=50.0,
        channels=1,
        nozzle_offset_z=-259.15,
        home_position=230.15,
        flow_rates=FlowRates(
            default_blow_out={"2.14": 57},
            default_aspirate={"2.14": 35},
            default_dispense={"2.14": 57},
        ),
        tip_configuration_lookup_table=result1.tip_configuration_lookup_table,
        nominal_tip_overlap=result1.nominal_tip_overlap,
        nozzle_map=result1.nozzle_map,
        back_left_corner_offset=Point(-38, 0.0, -259.15),
        front_right_corner_offset=Point(11.5, -64.0, -259.15),
        pipette_lld_settings={
            "t20": {"minHeight": 1.5, "minVolume": 0.0},
            "t50": {"minHeight": 1.0, "minVolume": 0.0},
        },
        plunger_positions={
            "top": 0.0,
            "bottom": 71.5,
            "blow_out": 76.5,
            "drop_tip": 90.5,
        },
        shaft_ul_per_mm=0.785,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={
            VolumeModes.default: 5.0,
            VolumeModes.lowVolumeDefault: 0.5,
        },
    )
    subject_instance.configure_virtual_pipette_for_volume(
        "my-pipette", 1, result1.model
    )
    result2 = subject_instance.get_virtual_pipette_static_config(
        PipetteNameType.P50_SINGLE_FLEX.value, "my-pipette", "v0"
    )
    assert result2 == LoadedStaticPipetteData(
        model="p50_single_v3.6",
        display_name="Flex 1-Channel 50 µL",
        min_volume=0.5,
        max_volume=30,
        channels=1,
        nozzle_offset_z=-259.15,
        home_position=230.15,
        flow_rates=FlowRates(
            default_blow_out={"2.14": 26.7},
            default_aspirate={"2.14": 26.7},
            default_dispense={"2.14": 26.7},
        ),
        tip_configuration_lookup_table=result2.tip_configuration_lookup_table,
        nominal_tip_overlap=result2.nominal_tip_overlap,
        nozzle_map=result2.nozzle_map,
        back_left_corner_offset=Point(-38, 0.0, -259.15),
        front_right_corner_offset=Point(11.5, -64.0, -259.15),
        pipette_lld_settings={
            "t20": {"minHeight": 1.5, "minVolume": 0.0},
            "t50": {"minHeight": 1.0, "minVolume": 0.0},
        },
        plunger_positions={
            "top": 0.0,
            "bottom": 61.5,
            "blow_out": 76.5,
            "drop_tip": 90.5,
        },
        shaft_ul_per_mm=0.785,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.lowVolumeDefault,
        available_volume_modes_min_vol={
            VolumeModes.default: 5.0,
            VolumeModes.lowVolumeDefault: 0.5,
        },
    )


def test_load_virtual_pipette_by_model_string(
    subject_instance: VirtualPipetteDataProvider,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should return config data given a pipette model."""
    result = subject_instance.get_virtual_pipette_static_config_by_model_string(
        "p300_multi_v2.1", "my-pipette", "v0"
    )
    assert result == LoadedStaticPipetteData(
        model="p300_multi_v2.1",
        display_name="P300 8-Channel GEN2",
        min_volume=20.0,
        max_volume=300,
        channels=8,
        nozzle_offset_z=35.52,
        home_position=155.75,
        flow_rates=FlowRates(
            default_blow_out={"2.0": 94.0},
            default_aspirate={"2.0": 94.0},
            default_dispense={"2.0": 94.0},
        ),
        tip_configuration_lookup_table=result.tip_configuration_lookup_table,
        nominal_tip_overlap=result.nominal_tip_overlap,
        nozzle_map=result.nozzle_map,
        back_left_corner_offset=Point(-16.0, 43.15, 35.52),
        front_right_corner_offset=Point(16.0, -43.15, 35.52),
        pipette_lld_settings={},
        plunger_positions={
            "top": 19.5,
            "bottom": -14.5,
            "blow_out": -19.0,
            "drop_tip": -33.4,
        },
        shaft_ul_per_mm=9.621,
        available_sensors=AvailableSensorDefinition(sensors=[]),
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={
            VolumeModes.default: 20.0,
        },
    )


def test_load_virtual_pipette_nozzle_layout(
    subject_instance: VirtualPipetteDataProvider,
) -> None:
    """It should return a NozzleMap object."""
    subject_instance.configure_virtual_pipette_nozzle_layout(
        "my-pipette", "p300_multi_v2.1", "D1", "H1", "H1"
    )
    result = subject_instance.get_nozzle_layout_for_pipette("my-pipette")
    assert result.configuration.value == "COLUMN"
    assert result.starting_nozzle == "H1"
    assert result.front_right == "H1"
    assert result.back_left == "D1"

    subject_instance.configure_virtual_pipette_nozzle_layout(
        "my-pipette", "p300_multi_v2.1"
    )
    result = subject_instance.get_nozzle_layout_for_pipette("my-pipette")
    assert result.configuration.value == "FULL"

    subject_instance.configure_virtual_pipette_nozzle_layout(
        "my-96-pipette", "p1000_96_v3.6", "A1", "A12", "A1"
    )
    result = subject_instance.get_nozzle_layout_for_pipette("my-96-pipette")
    assert result.configuration.value == "ROW"

    subject_instance.configure_virtual_pipette_nozzle_layout(
        "my-96-pipette", "p1000_96_v3.6", "A1", "A1"
    )
    result = subject_instance.get_nozzle_layout_for_pipette("my-96-pipette")
    assert result.configuration.value == "SINGLE"

    subject_instance.configure_virtual_pipette_nozzle_layout(
        "my-96-pipette", "p1000_96_v3.6", "A1", "H1"
    )
    result = subject_instance.get_nozzle_layout_for_pipette("my-96-pipette")
    assert result.configuration.value == "COLUMN"


@pytest.fixture
def pipette_dict(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: AvailableSensorDefinition,
) -> PipetteDict:
    """Get a pipette dict."""
    return {
        "name": "p300_single_gen2",
        "min_volume": 20,
        "max_volume": 300,
        "channels": 1,
        "aspirate_flow_rate": 46.43,
        "dispense_flow_rate": 46.43,
        "pipette_id": "P3HSV202020060308",
        "current_volume": 0.0,
        "display_name": "P300 Single-Channel GEN2",
        "tip_length": 0.0,
        "model": PipetteModel("p300_single_v2.0"),
        "blow_out_flow_rate": 46.43,
        "working_volume": 300,
        "tip_overlap": {
            "default": 8.2,
            "opentrons/opentrons_96_tiprack_300ul/1": 8.2,
            "opentrons/opentrons_96_filtertiprack_200ul/1": 8.2,
        },
        "versioned_tip_overlap": {
            "v0": {
                "default": 8.2,
            },
            "v2": {"default": 9.3},
        },
        "available_volume": 300.0,
        "return_tip_height": 0.5,
        "default_aspirate_flow_rates": {"2.0": 46.43, "2.1": 92.86},
        "default_blow_out_flow_rates": {"2.0": 46.43, "2.2": 92.86},
        "default_dispense_flow_rates": {"2.0": 46.43, "2.3": 92.86},
        "back_compat_names": ["p300_single"],
        "has_tip": False,
        "aspirate_speed": 5.021202,
        "dispense_speed": 5.021202,
        "blow_out_speed": 5.021202,
        "ready_to_aspirate": False,
        "default_blow_out_speeds": {"2.0": 5.021202, "2.6": 10.042404},
        "default_dispense_speeds": {"2.0": 5.021202, "2.6": 10.042404},
        "default_aspirate_speeds": {"2.0": 5.021202, "2.6": 10.042404},
        "default_push_out_volume": 3,
        "supported_tips": {pip_types.PipetteTipType.t300: supported_tip_fixture},
        "current_nozzle_map": get_default_nozzle_map(PipetteNameType.P300_SINGLE_GEN2),
        "pipette_bounding_box_offsets": PipetteBoundingBoxOffsetDefinition(
            backLeftCorner=[10, 20, 30],
            frontRightCorner=[40, 50, 60],
        ),
        "lld_settings": {
            "t50": {"minHeight": 0.5, "minVolume": 0},
            "t200": {"minHeight": 0.5, "minVolume": 0},
            "t1000": {"minHeight": 0.5, "minVolume": 0},
        },
        "plunger_positions": {"top": 100, "bottom": 20, "blow_out": 10, "drop_tip": 0},
        "shaft_ul_per_mm": 5.0,
        "available_sensors": available_sensors,
        "volume_mode": VolumeModes.lowVolumeDefault,
        "available_volume_modes": {},
    }


@pytest.mark.parametrize(
    "tip_overlap_version,overlap_data",
    [
        ("v0", {"default": 8.2}),
        ("v1", {"default": 8.2}),
        ("v2", {"default": 9.3}),
        ("v10000", {"default": 9.3}),
    ],
)
def test_get_pipette_static_config(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    pipette_dict: PipetteDict,
    tip_overlap_version: str,
    overlap_data: Dict[str, float],
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should return config data given a PipetteDict."""
    result = subject.get_pipette_static_config(pipette_dict, tip_overlap_version)

    assert result == LoadedStaticPipetteData(
        model="p300_single_v2.0",
        display_name="P300 Single-Channel GEN2",
        min_volume=20,
        max_volume=300,
        channels=1,
        flow_rates=FlowRates(
            default_aspirate={"2.0": 46.43, "2.1": 92.86},
            default_dispense={"2.0": 46.43, "2.3": 92.86},
            default_blow_out={"2.0": 46.43, "2.2": 92.86},
        ),
        tip_configuration_lookup_table={300: supported_tip_fixture},
        nominal_tip_overlap=overlap_data,
        # TODO(mc, 2023-02-28): these two values are not present in PipetteDict
        # https://opentrons.atlassian.net/browse/RCORE-655
        nozzle_offset_z=0,
        home_position=0,
        nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE_GEN2),
        back_left_corner_offset=Point(10, 20, 30),
        front_right_corner_offset=Point(40, 50, 60),
        pipette_lld_settings={
            "t50": {"minHeight": 0.5, "minVolume": 0},
            "t200": {"minHeight": 0.5, "minVolume": 0},
            "t1000": {"minHeight": 0.5, "minVolume": 0},
        },
        plunger_positions={"top": 100, "bottom": 20, "blow_out": 10, "drop_tip": 0},
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.lowVolumeDefault,
        available_volume_modes_min_vol={},
    )


@pytest.mark.parametrize(
    "version",
    [
        "",
        "qwe",
        "v",
        "v-1",
        "vab",
    ],
)
def test_validate_bad_tip_overlap_versions(version: str) -> None:
    """Raise for bad tip overlap version specs."""
    with pytest.raises(InvalidLoadPipetteSpecsError):
        validate_and_default_tip_overlap_version(version)


def test_default_tip_overlap_versions() -> None:
    """Default None tip overlap version specs."""
    assert (
        validate_and_default_tip_overlap_version(None)
        == f"v{TIP_OVERLAP_VERSION_MAXIMUM}"
    )


@pytest.mark.parametrize("version", ["v0", "v1", f"v{maxsize + 1}"])
def test_pass_valid_tip_overlap_versions(version: str) -> None:
    """Pass valid tip overlap specs."""
    assert validate_and_default_tip_overlap_version(version) == version


@pytest.mark.parametrize(
    "version,target_data",
    [
        ("v0", {"default": 123.0}),
        ("v1", {"default": 321.1}),
        ("v3", {"default": 333.5}),
        ("v9999", {"default": 4414.99}),
    ],
)
def test_get_latest_tip_overlap(version: str, target_data: Dict[str, float]) -> None:
    """Test the search function for latest offset."""
    overlap = {
        "v0": {"default": 123.0},
        "v1": {"default": 321.1},
        "v2": {"default": 333.5},
        "v1231": {"default": 4414.99},
    }
    assert get_latest_tip_overlap_before_version(overlap, version) == target_data



"""Test gantry movement handler with hardware API."""

from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, Optional

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import PositionUnknownError

from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.types import (
    Axis as HardwareAxis,
)
from opentrons.hardware_control.types import (
    CriticalPoint,
)
from opentrons.motion_planning import Waypoint
from opentrons.protocol_engine.errors import InvalidAxisForRobotType, MustHomeError
from opentrons.protocol_engine.execution.gantry_mover import (
    VIRTUAL_MAX_OT3_HEIGHT,
    HardwareGantryMover,
    VirtualGantryMover,
    create_gantry_mover,
)
from opentrons.protocol_engine.state.motion import PipetteLocationData
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    CurrentWell,
    DeckPoint,
    MotorAxis,
    TipGeometry,
)
from opentrons.types import Mount, MountType, Point

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture
def mock_hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mock in the shape of a HardwareAPI."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock in the shape of a StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def hardware_subject(
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
) -> HardwareGantryMover:
    """Create a GantryMover with its dependencies mocked out."""
    return HardwareGantryMover(
        hardware_api=mock_hardware_api,
        state_view=mock_state_view,
    )


@pytest.fixture
def virtual_subject(
    mock_state_view: StateView,
) -> VirtualGantryMover:
    """Create a GantryMover with its dependencies mocked out."""
    return VirtualGantryMover(state_view=mock_state_view)


async def test_create_gantry_movement_handler(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
) -> None:
    """It should return virtual or real gantry movement handlers depending on config."""
    decoy.when(mock_state_view.config.use_virtual_pipettes).then_return(False)
    assert isinstance(
        create_gantry_mover(
            state_view=mock_state_view,
            hardware_api=mock_hardware_api,
        ),
        HardwareGantryMover,
    )

    decoy.when(mock_state_view.config.use_virtual_pipettes).then_return(True)
    assert isinstance(
        create_gantry_mover(
            state_view=mock_state_view,
            hardware_api=mock_hardware_api,
        ),
        VirtualGantryMover,
    )


async def test_get_position(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should get the position of the pipette with the hardware API."""
    current_well = CurrentWell(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
    )
    decoy.when(
        mock_state_view.motion.get_pipette_location("pipette-id", current_well)
    ).then_return(
        PipetteLocationData(
            mount=MountType.RIGHT,
            critical_point=CriticalPoint.XY_CENTER,
        )
    )
    decoy.when(
        await mock_hardware_api.gantry_position(
            mount=Mount.RIGHT,
            critical_point=CriticalPoint.XY_CENTER,
            fail_on_not_homed=True,
        )
    ).then_return(Point(1, 2, 3))

    result = await hardware_subject.get_position(
        "pipette-id", current_well=current_well, fail_on_not_homed=True
    )

    assert result == Point(1, 2, 3)


async def test_get_position_raises(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should raise a MustHomeError."""
    decoy.when(
        mock_state_view.motion.get_pipette_location("pipette-id", None)
    ).then_return(
        PipetteLocationData(
            mount=MountType.LEFT,
            critical_point=CriticalPoint.NOZZLE,
        )
    )
    decoy.when(
        await mock_hardware_api.gantry_position(
            mount=Mount.LEFT,
            critical_point=CriticalPoint.NOZZLE,
            fail_on_not_homed=False,
        )
    ).then_raise(PositionUnknownError("oh no"))

    with pytest.raises(MustHomeError, match="oh no"):
        await hardware_subject.get_position("pipette-id")


def test_get_max_travel_z(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should get the max travel z height with the hardware API."""
    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.RIGHT
    )
    decoy.when(
        mock_hardware_api.get_instrument_max_height(mount=Mount.RIGHT)
    ).then_return(42.1)

    assert hardware_subject.get_max_travel_z("pipette-id") == 42.1


async def test_move_to(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should move the gantry with the hardware API."""
    decoy.when(mock_state_view.pipettes.get_mount("abc123")).then_return(
        MountType.RIGHT
    )

    result = await hardware_subject.move_to(
        pipette_id="abc123",
        waypoints=[
            Waypoint(position=Point(1, 2, 3), critical_point=CriticalPoint.TIP),
            Waypoint(position=Point(4, 5, 6), critical_point=CriticalPoint.XY_CENTER),
        ],
        speed=9001,
    )

    assert result == Point(4, 5, 6)

    decoy.verify(
        await mock_hardware_api.move_to(
            mount=Mount.RIGHT,
            abs_position=Point(1, 2, 3),
            critical_point=CriticalPoint.TIP,
            speed=9001,
        ),
        await mock_hardware_api.move_to(
            mount=Mount.RIGHT,
            abs_position=Point(4, 5, 6),
            critical_point=CriticalPoint.XY_CENTER,
            speed=9001,
        ),
    )


async def test_move_relative(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should move the gantry by the delta with the hardware API."""
    decoy.when(mock_state_view.motion.get_pipette_location("pipette-id")).then_return(
        PipetteLocationData(
            mount=MountType.RIGHT,
            critical_point=CriticalPoint.XY_CENTER,
        )
    )
    decoy.when(
        await mock_hardware_api.gantry_position(
            mount=Mount.RIGHT,
            critical_point=CriticalPoint.XY_CENTER,
            fail_on_not_homed=True,
        )
    ).then_return(Point(4, 5, 6))

    result = await hardware_subject.move_relative(
        pipette_id="pipette-id",
        delta=Point(1, 2, 3),
        speed=9001,
    )

    assert result == Point(4, 5, 6)

    # TODO(mc, 2022-05-13): the order of these calls is difficult to manage
    # and test for. Ideally, `hardware.move_rel` would return the resulting position
    decoy.verify(
        await mock_hardware_api.move_rel(
            mount=Mount.RIGHT,
            delta=Point(1, 2, 3),
            fail_on_not_homed=True,
            speed=9001,
        ),
        times=1,
    )


async def test_move_relative_must_home(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    mock_state_view: StateView,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should raise a MustHomeError."""
    decoy.when(mock_state_view.motion.get_pipette_location("pipette-id")).then_return(
        PipetteLocationData(
            mount=MountType.LEFT,
            critical_point=CriticalPoint.XY_CENTER,
        )
    )
    decoy.when(
        await mock_hardware_api.move_rel(  # type: ignore[func-returns-value]
            mount=Mount.LEFT,
            delta=Point(x=1, y=2, z=3),
            fail_on_not_homed=True,
            speed=456.7,
        )
    ).then_raise(PositionUnknownError("oh no"))

    with pytest.raises(MustHomeError, match="oh no"):
        await hardware_subject.move_relative(
            pipette_id="pipette-id",
            delta=Point(x=1, y=2, z=3),
            speed=456.7,
        )


async def test_home(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwareGantryMover,
    mock_state_view: StateView,
) -> None:
    """It should home a set of axes."""
    decoy.when(mock_state_view.config.robot_type).then_return("OT-2 Standard")
    await hardware_subject.home(
        axes=[
            MotorAxis.X,
            MotorAxis.Y,
            MotorAxis.LEFT_Z,
            MotorAxis.RIGHT_Z,
            MotorAxis.LEFT_PLUNGER,
            MotorAxis.RIGHT_PLUNGER,
        ]
    )
    decoy.verify(
        await mock_hardware_api.home(
            axes=[
                HardwareAxis.X,
                HardwareAxis.Y,
                HardwareAxis.Z,
                HardwareAxis.A,
                HardwareAxis.B,
                HardwareAxis.C,
            ]
        ),
        times=1,
    )
    decoy.reset()

    await hardware_subject.home(axes=None)
    decoy.verify(await mock_hardware_api.home(), times=1)
    decoy.reset()

    await hardware_subject.home(axes=[])
    decoy.verify(await mock_hardware_api.home(axes=[]), times=1)


async def test_ot2_home_fails_with_ot3_axes(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwareGantryMover,
    mock_state_view: StateView,
) -> None:
    """It should raise an error when homing axes that don't exist on OT2."""
    decoy.when(mock_state_view.config.robot_type).then_return("OT-2 Standard")
    with pytest.raises(InvalidAxisForRobotType):
        await hardware_subject.home(
            axes=[
                MotorAxis.LEFT_PLUNGER,
                MotorAxis.RIGHT_PLUNGER,
                MotorAxis.EXTENSION_Z,
                MotorAxis.EXTENSION_JAW,
            ]
        )


@pytest.mark.ot3_only
async def test_home_on_ot3(
    decoy: Decoy,
    ot3_hardware_api: OT3API,
    mock_state_view: StateView,
) -> None:
    """Test homing all OT3 axes."""
    subject = HardwareGantryMover(
        state_view=mock_state_view, hardware_api=ot3_hardware_api
    )
    decoy.when(mock_state_view.config.robot_type).then_return("OT-3 Standard")
    await subject.home(
        axes=[
            MotorAxis.X,
            MotorAxis.Y,
            MotorAxis.LEFT_Z,
            MotorAxis.RIGHT_Z,
            MotorAxis.LEFT_PLUNGER,
            MotorAxis.RIGHT_PLUNGER,
            MotorAxis.EXTENSION_JAW,
            MotorAxis.EXTENSION_Z,
        ]
    )
    decoy.verify(
        await ot3_hardware_api.home(
            axes=[
                HardwareAxis.X,
                HardwareAxis.Y,
                HardwareAxis.Z,
                HardwareAxis.A,
                HardwareAxis.B,
                HardwareAxis.C,
                HardwareAxis.G,
                HardwareAxis.Z_G,
            ]
        ),
    )


async def test_retract_axis(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwareGantryMover,
    mock_state_view: StateView,
) -> None:
    """It should send a hardware control retract axis command with specified axis."""
    decoy.when(mock_state_view.config.robot_type).then_return("OT-2 Standard")
    await hardware_subject.retract_axis(axis=MotorAxis.RIGHT_Z)
    decoy.verify(
        await mock_hardware_api.retract_axis(axis=HardwareAxis.A),
        times=1,
    )


async def test_retract_axis_with_invalid_axis_for_ot2(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwareGantryMover,
    mock_state_view: StateView,
) -> None:
    """It should raise error when trying to retract an axis that's not valid on OT2."""
    decoy.when(mock_state_view.config.robot_type).then_return("OT-2 Standard")
    with pytest.raises(InvalidAxisForRobotType):
        await hardware_subject.retract_axis(axis=MotorAxis.EXTENSION_Z)


@pytest.mark.ot3_only
async def test_retract_axis_on_ot3(
    decoy: Decoy,
    ot3_hardware_api: OT3API,
    mock_state_view: StateView,
) -> None:
    """It should call OT3 hardware API's retract axis with specified axis."""
    subject = HardwareGantryMover(
        state_view=mock_state_view, hardware_api=ot3_hardware_api
    )
    decoy.when(mock_state_view.config.robot_type).then_return("OT-3 Standard")
    await subject.retract_axis(MotorAxis.EXTENSION_Z)
    decoy.verify(await ot3_hardware_api.retract_axis(axis=HardwareAxis.Z_G), times=1)


# TODO(mc, 2022-12-01): this is overly complicated
# https://opentrons.atlassian.net/browse/RET-1287
async def test_home_z(
    decoy: Decoy,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwareGantryMover,
) -> None:
    """It should home a single Z axis and plunger."""
    await hardware_subject.home(axes=[MotorAxis.LEFT_Z, MotorAxis.LEFT_PLUNGER])
    decoy.verify(
        await mock_hardware_api.home_z(Mount.LEFT),
        await mock_hardware_api.home_plunger(Mount.LEFT),
    )
    decoy.reset()

    await hardware_subject.home(axes=[MotorAxis.RIGHT_Z, MotorAxis.RIGHT_PLUNGER])
    decoy.verify(
        await mock_hardware_api.home_z(Mount.RIGHT),
        await mock_hardware_api.home_plunger(Mount.RIGHT),
    )
    decoy.reset()

    await hardware_subject.home(axes=[MotorAxis.LEFT_PLUNGER])
    decoy.verify(
        await mock_hardware_api.home_plunger(Mount.LEFT),
        times=1,
    )
    decoy.reset()

    await hardware_subject.home(axes=[MotorAxis.RIGHT_PLUNGER])
    decoy.verify(
        await mock_hardware_api.home_plunger(Mount.RIGHT),
        times=1,
    )
    decoy.reset()

    await hardware_subject.home(axes=[MotorAxis.RIGHT_Z, MotorAxis.LEFT_PLUNGER])
    decoy.verify(
        await mock_hardware_api.home([HardwareAxis.A, HardwareAxis.B]),
        times=1,
    )


@pytest.mark.parametrize(
    argnames=[
        "axis_map",
        "critical_point",
        "relative_move",
        "expected_mount",
        "call_to_hw",
        "final_position",
    ],
    argvalues=[
        [
            {MotorAxis.X: 10.0, MotorAxis.Y: 15.0, MotorAxis.RIGHT_Z: 20.0},
            {MotorAxis.X: 2.0, MotorAxis.Y: 1.0, MotorAxis.RIGHT_Z: 1.0},
            False,
            Mount.RIGHT,
            OrderedDict(
                {HardwareAxis.X: -2.0, HardwareAxis.Y: 4.0, HardwareAxis.A: 19.0}
            ),
            {HardwareAxis.X: -2.0, HardwareAxis.Y: 4.0, HardwareAxis.A: 9.0},
        ],
        [
            {MotorAxis.RIGHT_Z: 20.0},
            None,
            True,
            Mount.RIGHT,
            OrderedDict({HardwareAxis.A: 29.0}),
            {
                HardwareAxis.X: 10.0,
                HardwareAxis.Y: 15.0,
                HardwareAxis.Z: 10.0,
                HardwareAxis.A: 30.0,
            },
        ],
        [
            {MotorAxis.AXIS_96_CHANNEL_CAM: 10.0},
            None,
            False,
            Mount.LEFT,
            OrderedDict({HardwareAxis.Q: 10.0}),
            {HardwareAxis.Q: 10.0},
        ],
    ],
)
@pytest.mark.ot3_only
async def test_move_axes(
    decoy: Decoy,
    ot3_hardware_api: OT3API,
    mock_state_view: StateView,
    axis_map: Dict[MotorAxis, float],
    critical_point: Optional[Dict[MotorAxis, float]],
    expected_mount: Mount,
    relative_move: bool,
    call_to_hw: "OrderedDict[HardwareAxis, float]",
    final_position: Dict[HardwareAxis, float],
) -> None:
    """Test the move axes function."""
    subject = HardwareGantryMover(
        state_view=mock_state_view, hardware_api=ot3_hardware_api
    )
    curr_pos = {
        HardwareAxis.X: 10.0,
        HardwareAxis.Y: 15.0,
        HardwareAxis.Z: 10.0,
        HardwareAxis.A: 10.0,
    }
    call_count = 0

    def _current_position(mount: Mount, refresh: bool) -> Dict[HardwareAxis, float]:
        nonlocal call_count
        nonlocal curr_pos
        nonlocal final_position
        if call_count == 0 and relative_move:
            call_count += 1
            return curr_pos
        else:
            return final_position

    decoy.when(
        await ot3_hardware_api.current_position(expected_mount, refresh=True)
    ).then_do(_current_position)

    decoy.when(ot3_hardware_api.config.left_mount_offset).then_return(Point(1, 1, 1))
    decoy.when(ot3_hardware_api.config.right_mount_offset).then_return(
        Point(10, 10, 10)
    )
    decoy.when(ot3_hardware_api.config.gripper_mount_offset).then_return(
        Point(0.5, 0.5, 0.5)
    )

    decoy.when(ot3_hardware_api.get_deck_from_machine(curr_pos)).then_return(curr_pos)

    decoy.when(ot3_hardware_api.get_deck_from_machine(final_position)).then_return(
        final_position
    )
    if not critical_point:
        decoy.when(ot3_hardware_api.critical_point_for(expected_mount)).then_return(
            Point(1, 1, 1)
        )

    pos = await subject.move_axes(axis_map, critical_point, 100, relative_move)
    decoy.verify(
        await ot3_hardware_api.move_axes(
            position=call_to_hw, speed=100, expect_stalls=False
        ),
        times=1,
    )
    assert pos == {
        subject._hardware_axis_to_motor_axis(ax): pos
        for ax, pos in final_position.items()
    }


async def test_virtual_get_position(
    decoy: Decoy,
    mock_state_view: StateView,
    virtual_subject: VirtualGantryMover,
) -> None:
    """It should get the position of the pipette with the state store."""
    decoy.when(mock_state_view.pipettes.get_deck_point("pipette-id")).then_return(
        DeckPoint(x=1, y=2, z=3)
    )

    result = await virtual_subject.get_position("pipette-id")

    assert result == Point(x=1, y=2, z=3)


async def test_virtual_get_position_default(
    decoy: Decoy,
    mock_state_view: StateView,
    virtual_subject: VirtualGantryMover,
) -> None:
    """It should get a default Point if no stored deck point can be found in the state store."""
    decoy.when(mock_state_view.pipettes.get_deck_point("pipette-id")).then_return(None)

    result = await virtual_subject.get_position("pipette-id")

    assert result == Point(x=0, y=0, z=0)


def test_virtual_get_max_travel_z_ot2(
    decoy: Decoy,
    mock_state_view: StateView,
    virtual_subject: VirtualGantryMover,
) -> None:
    """It should get the max travel z height with the state store for an OT-2."""
    decoy.when(mock_state_view.config.robot_type).then_return("OT-2 Standard")
    decoy.when(
        mock_state_view.pipettes.get_instrument_max_height_ot2("pipette-id")
    ).then_return(42)
    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=20, diameter=0, volume=0)
    )

    result = virtual_subject.get_max_travel_z("pipette-id")

    assert result == 22.0


def test_virtual_get_max_travel_z_ot3(
    decoy: Decoy,
    mock_state_view: StateView,
    virtual_subject: VirtualGantryMover,
) -> None:
    """It should get the max travel z height with the state store."""
    decoy.when(mock_state_view.config.robot_type).then_return("OT-3 Standard")
    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=48, diameter=0, volume=0)
    )

    result = virtual_subject.get_max_travel_z("pipette-id")

    assert result == VIRTUAL_MAX_OT3_HEIGHT - 48.0


async def test_virtual_move_relative(
    decoy: Decoy,
    mock_state_view: StateView,
    virtual_subject: VirtualGantryMover,
) -> None:
    """It should simulate moving the gantry by the delta with the state store."""
    decoy.when(mock_state_view.pipettes.get_deck_point("pipette-id")).then_return(
        DeckPoint(x=1, y=2, z=3)
    )

    result = await virtual_subject.move_relative(
        "pipette-id",
        delta=Point(3, 2, 1),
        speed=123,
    )

    assert result == Point(x=4, y=4, z=4)


async def test_virtual_move_to(
    decoy: Decoy, virtual_subject: VirtualGantryMover
) -> None:
    """It should no-op on move to, returning the last waypoint."""
    result = await virtual_subject.move_to(
        pipette_id="abc123",
        waypoints=[
            Waypoint(position=Point(1, 2, 3), critical_point=CriticalPoint.TIP),
            Waypoint(position=Point(4, 5, 6), critical_point=CriticalPoint.XY_CENTER),
        ],
        speed=None,
    )

    assert result == Point(4, 5, 6)


@pytest.mark.parametrize(
    argnames=["axis_map", "critical_point", "relative_move", "expected_position"],
    argvalues=[
        [
            {MotorAxis.X: 10, MotorAxis.Y: 15, MotorAxis.RIGHT_Z: 20},
            {MotorAxis.X: 2, MotorAxis.Y: 1, MotorAxis.RIGHT_Z: 1},
            False,
            {MotorAxis.X: 8, MotorAxis.Y: 14, MotorAxis.RIGHT_Z: 19},
        ],
        [
            {MotorAxis.RIGHT_Z: 20},
            None,
            True,
            {MotorAxis.X: 0.0, MotorAxis.Y: 0.0, MotorAxis.RIGHT_Z: 20},
        ],
        [
            {MotorAxis.AXIS_96_CHANNEL_CAM: 10},
            None,
            False,
            {MotorAxis.AXIS_96_CHANNEL_CAM: 10},
        ],
    ],
)
async def test_virtual_move_axes(
    decoy: Decoy,
    virtual_subject: VirtualGantryMover,
    axis_map: Dict[MotorAxis, float],
    critical_point: Optional[Dict[MotorAxis, float]],
    relative_move: bool,
    expected_position: Dict[MotorAxis, float],
) -> None:
    """It should simulate moving a set of axis by a certain distance."""
    pos = await virtual_subject.move_axes(axis_map, critical_point, 100, relative_move)
    assert pos == expected_position



import json
import typing
from pathlib import Path

import pydantic
import pytest

from opentrons.calibration_storage import file_operators as io


class DummyModel(pydantic.BaseModel):
    integer_field: int
    aliased_field: str = pydantic.Field(alias="! aliased field !")


@pytest.fixture
def calibration() -> typing.Dict[str, typing.Any]:
    return {
        "attitude": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        "lastModified": "2022-10-27T15:21:43.195599+00:00",
        "source": "user",
        "pipette_calibrated_with": "pip1",
        "tiprack": "mytiprack",
        "status": {"markedBad": False, "source": None, "markedAt": None},
    }


@pytest.fixture
def malformed_calibration() -> typing.Dict[str, typing.Any]:
    return {
        "attitude": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        "lastModified": "blah",
        "source": "user",
        "pipette_calibrated_with": "pip1",
        "tiprack": "mytiprack",
        "status": {"markedBad": False, "source": None, "markedAt": None},
    }


def test_save_delete_file(
    tmp_path: Path, calibration: typing.Dict[str, typing.Any]
) -> None:
    """
    Test saving, reading and deleting a correctly formed calibration blob.
    """
    my_calpath = tmp_path / "calibrations"
    io.save_to_file(my_calpath, "my_calibration", calibration)
    file_to_delete = my_calpath / "my_calibration.json"
    assert io.read_cal_file(file_to_delete)

    io.delete_file(file_to_delete)

    with pytest.raises(FileNotFoundError):
        io.read_cal_file(file_to_delete)


def test_malformed_calibration(
    tmp_path: Path, malformed_calibration: typing.Dict[str, typing.Any]
) -> None:
    """
    Test saving and reading a malformed calibration blob. Note, malformed here only means
    datetime abnormalities.
    """
    malformed_calibration_dir = tmp_path / "calibrations"
    malformed_calibration_path = malformed_calibration_dir / "my_bad_calibration.json"

    # TODO (lc 10-27-2022) We don't actually throw an error when we're saving bad calibration data.
    # Probably before this point, we should make sure that we're passin in a validated pydantic
    # model because otherwise we could potentially be saving malformed data which would fail on
    # a read file.
    io.save_to_file(
        malformed_calibration_dir, "my_bad_calibration", malformed_calibration
    )

    malformed_calibration_dir.mkdir(parents=True, exist_ok=True)
    malformed_calibration_path.write_text(
        json.dumps(malformed_calibration), encoding="utf-8"
    )
    with pytest.raises(AssertionError):
        io.read_cal_file(malformed_calibration_path)


def test_deserialize_pydantic_model_valid() -> None:
    serialized = b'{"integer_field": 123, "! aliased field !": "abc"}'
    assert io.deserialize_pydantic_model(
        serialized, DummyModel
    ) == DummyModel.model_construct(integer_field=123, aliased_field="abc")


def test_deserialize_pydantic_model_invalid_as_json() -> None:
    serialized = "😾".encode("utf-8")
    assert io.deserialize_pydantic_model(serialized, DummyModel) is None
    # Ideally we would assert that the subject logged a message saying "not valid JSON",
    # but the opentrons.simulate and opentrons.execute tests interfere with the process's logger
    # settings and prevent that message from showing up in pytest's caplog fixture.


def test_read_pydantic_model_from_file_invalid_model(tmp_path: Path) -> None:
    serialized = b'{"integer_field": "not an integer"}'
    assert io.deserialize_pydantic_model(serialized, DummyModel) is None
    # Ideally we would assert that the subject logged a message saying "does not match model",
    # but the opentrons.simulate and opentrons.execute tests interfere with the process's logger
    # settings and prevent that message from showing up in pytest's caplog fixture.



from typing import TYPE_CHECKING, Generator
from unittest.mock import MagicMock, patch

import pytest

from opentrons_shared_data.robot.types import RobotTypeEnum

from opentrons.config import reset

if TYPE_CHECKING:
    from opentrons_shared_data.deck.types import RobotModel


@pytest.fixture
def robot_type_enum(robot_model: "RobotModel") -> RobotTypeEnum:
    return RobotTypeEnum.OT2 if robot_model == "OT-2 Standard" else RobotTypeEnum.FLEX


@pytest.fixture
def mock_reset_boot_scripts() -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.reset_boot_scripts") as m:
        yield m


@pytest.fixture
def mock_reset_pipette_offset() -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.reset_pipette_offset") as m:
        yield m


@pytest.fixture
def mock_reset_deck_calibration() -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.reset_deck_calibration") as m:
        yield m


@pytest.fixture
def mock_reset_tip_length_calibrations() -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.reset_tip_length_calibrations") as m:
        yield m


@pytest.fixture
def mock_cal_pipette_offset(
    robot_model: "RobotModel",
) -> Generator[MagicMock, None, None]:
    prefix = "ot2" if robot_model == "OT-2 Standard" else "ot3"
    with patch(
        f"opentrons.config.reset.{prefix}.clear_pipette_offset_calibrations"
    ) as m:
        yield m


@pytest.fixture
def mock_cal_gripper_offset(
    robot_model: "RobotModel",
) -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.reset_gripper_offset") as m:
        yield m


@pytest.fixture
def mock_cal_tip_length(robot_model: "RobotModel") -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.ot2.clear_tip_length_calibration") as m:
        yield m


@pytest.fixture
def mock_cal_robot_attitude(
    robot_model: "RobotModel",
) -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.delete_robot_deck_attitude") as m:
        yield m


@pytest.fixture
def mock_cal_module_offset(
    robot_model: "RobotModel",
) -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.reset.reset_module_calibration") as m:
        yield m


def test_get_options() -> None:
    options = reset.reset_options(RobotTypeEnum.OT2)
    assert list(options.keys()) == reset._OT_2_RESET_OPTIONS

    options = reset.reset_options(RobotTypeEnum.FLEX)
    assert list(options.keys()) == reset._FLEX_RESET_OPTIONS


def test_reset_empty_set(
    mock_reset_boot_scripts: MagicMock,
    mock_reset_pipette_offset: MagicMock,
    mock_reset_deck_calibration: MagicMock,
    mock_reset_tip_length_calibrations: MagicMock,
    mock_cal_module_offset: MagicMock,
    robot_type_enum: RobotTypeEnum,
) -> None:
    reset.reset(set(), robot_type_enum)
    mock_reset_boot_scripts.assert_not_called()
    mock_reset_pipette_offset.assert_not_called()
    mock_reset_deck_calibration.assert_not_called()
    mock_reset_tip_length_calibrations.assert_not_called()
    mock_cal_module_offset.assert_not_called()


def test_reset_all_set(
    mock_reset_boot_scripts: MagicMock,
    mock_reset_pipette_offset: MagicMock,
    mock_reset_deck_calibration: MagicMock,
    mock_reset_tip_length_calibrations: MagicMock,
    robot_type_enum: RobotTypeEnum,
) -> None:
    reset.reset(
        {
            reset.ResetOptionId.boot_scripts,
            reset.ResetOptionId.deck_calibration,
            reset.ResetOptionId.pipette_offset,
            reset.ResetOptionId.tip_length_calibrations,
        },
        robot_type_enum,
    )
    mock_reset_boot_scripts.assert_called_once()
    mock_reset_pipette_offset.assert_called_once()
    mock_reset_deck_calibration.assert_called_once()
    mock_reset_deck_calibration.assert_called_once()
    mock_reset_tip_length_calibrations.assert_called_once()


def test_deck_calibration_reset(
    mock_cal_pipette_offset: MagicMock,
    mock_cal_robot_attitude: MagicMock,
    robot_type_enum: RobotTypeEnum,
) -> None:
    reset.reset_deck_calibration(robot_type_enum)
    mock_cal_robot_attitude.assert_called_once()
    mock_cal_pipette_offset.assert_called_once()


@pytest.mark.ot2_only
def test_tip_length_calibrations_reset(
    mock_cal_pipette_offset: MagicMock,
    mock_cal_tip_length: MagicMock,
) -> None:
    reset.reset_tip_length_calibrations(RobotTypeEnum.OT2)
    mock_cal_tip_length.assert_called_once()
    mock_cal_pipette_offset.assert_called_once()


@pytest.mark.ot3_only
def test_tip_length_calibrations_reset_ot3() -> None:
    with pytest.raises(reset.UnrecognizedOption):
        reset.reset_tip_length_calibrations(RobotTypeEnum.FLEX)


def test_pipette_offset_reset(
    mock_cal_pipette_offset: MagicMock,
    robot_type_enum: RobotTypeEnum,
) -> None:
    reset.reset_pipette_offset(robot_type_enum)
    mock_cal_pipette_offset.assert_called_once()


def test_gripper_offset_reset(mock_cal_gripper_offset: MagicMock) -> None:
    reset.reset_gripper_offset()
    mock_cal_gripper_offset.assert_called_once()
